import site
import sys
from pathlib import Path
from typing import cast

from qtpy.QtGui import QTextDocument
from qtpy.QtWidgets import QGridLayout, QLabel, QMessageBox, QTabWidget, QTextBrowser, QWidget, QWidgetItem

__all__ = ["AboutBox", "about"]


class AboutBox(QMessageBox):
    def __init__(self, parent: QWidget | None = None, title: str = "", text: str = "") -> None:
        super().__init__(parent)

        if text or title:
            self.setText(text)
            self.setWindowTitle(title)

        self.setIcon(QMessageBox.Icon.Information)

        layout: QGridLayout = self.layout()

        def find_widget(name: str) -> QWidget | None:
            for i in range(layout.count()):
                widget: QWidget | None = cast(QWidgetItem, layout.itemAt(i)).widget()
                if widget is not None and widget.objectName() == name:
                    return widget

        icon_label: QLabel | None = find_widget("qt_msgboxex_icon_label")
        text_label: QLabel | None = find_widget("qt_msgbox_label")

        if parent is not None and icon_label is not None:
            icon_label.setPixmap(parent.windowIcon().pixmap(icon_label.size()))

        if text_label is not None:
            from ..utils import p_tag, tag

            text_label.hide()

            about_text: QTextBrowser = QTextBrowser(self)
            about_text.setText(text)
            cast(QTextDocument, about_text.document()).adjustSize()
            about_text.setMinimumSize(cast(QTextDocument, about_text.document()).size().toSize())

            tabs: QTabWidget = QTabWidget(self)
            tabs.setTabBarAutoHide(True)
            tabs.setTabPosition(QTabWidget.TabPosition.South)
            tabs.addTab(about_text, self.tr("About"))

            third_party_modules: list[str] = []
            prefixes: list[Path] = [
                Path(prefix).resolve() for prefix in site.getsitepackages([sys.exec_prefix, sys.prefix])
            ]
            for module_name, module in sys.modules.copy().items():
                paths = getattr(module, "__path__", [])
                if (
                    "." not in module_name
                    and module_name != "_distutils_hack"
                    and paths
                    and getattr(module, "__package__", "")
                    and any(prefix in Path(p).resolve().parents for p in paths for prefix in prefixes)
                ):
                    third_party_modules.append(module_name)
            if third_party_modules:
                lines: list[str] = [
                    self.tr("The app uses the following third-party modules:"),
                    tag(
                        "ul",
                        "".join(
                            map(
                                lambda s: tag("li", tag("tt", s)),
                                sorted(third_party_modules, key=str.casefold),
                            )
                        ),
                    ),
                ]
                third_party_label: QTextBrowser = QTextBrowser(self)
                third_party_label.setText(tag("html", "".join(map(p_tag, lines))))
                tabs.addTab(third_party_label, "Third-Party")
            layout.addWidget(tabs, 0, 2, 1, 1)


def about(parent: QWidget | None = None, title: str = "", text: str = "") -> int:
    box: AboutBox = AboutBox(parent=parent, title=title, text=text)
    return box.exec()



import enum
from math import floor, log10

from qtpy.QtCore import QLocale
from qtpy.QtGui import QValidator
from qtpy.QtWidgets import QDoubleSpinBox, QWidget

__all__ = ["FloatSpinBox"]


class FloatSpinBox(QDoubleSpinBox):
    class Mode(enum.Enum):
        auto = enum.auto()
        fixed = enum.auto()
        scientific = enum.auto()

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._mode: FloatSpinBox.Mode = FloatSpinBox.Mode.auto
        self._decimals: int = 2
        super().setDecimals(1000)

    @property
    def mode(self) -> Mode:
        return self._mode

    @mode.setter
    def mode(self, new_mode: "FloatSpinBox.Mode") -> None:
        if not isinstance(new_mode, FloatSpinBox.Mode):
            raise TypeError(f"Invalid mode: {new_mode}")
        self._mode = new_mode

    def valueFromText(self, text: str) -> float:
        locale: QLocale = QLocale()
        value: float
        ok: bool
        value, ok = locale.toDouble(text[len(self.prefix()) : -len(self.suffix())])
        if not ok:
            raise ValueError(f"could not convert string to float: {text[len(self.prefix()) : -len(self.suffix())]!r}")
        return value

    def textFromValue(self, v: float) -> str:
        locale: QLocale = QLocale()
        decimal_point: str = locale.decimalPoint()
        if self._mode == FloatSpinBox.Mode.auto:
            if abs(v) < self.singleStep() and v != 0.0:
                return f"{v:.{self._decimals}e}".replace(".", decimal_point)
            else:
                return f"{v:.{self._decimals}f}".replace(".", decimal_point)
        elif self._mode == FloatSpinBox.Mode.fixed:
            return f"{v:.{self._decimals}f}".replace(".", decimal_point)
        elif self._mode == FloatSpinBox.Mode.scientific:
            return f"{v:.{self._decimals}e}".replace(".", decimal_point)
        else:
            raise RuntimeError(f"Unknown mode {self._mode}")

    def validate(self, text: str, pos: int) -> tuple[QValidator.State, str, int]:
        try:
            self.valueFromText(text)
        except (ValueError, TypeError):
            return QValidator.State.Invalid, text, pos
        else:
            return QValidator.State.Acceptable, text, pos

    def fixup(self, text: str) -> str:
        for word in text.split():
            try:
                float(word)
            except ValueError:
                continue
            else:
                return word
        return ""

    def stepBy(self, steps: int) -> None:
        if self.value() != 0.0:
            exp: int = round(floor(log10(abs(self.value()))))
            self.setValue(self.value() + self.singleStep() * steps * 10.0**exp)
        else:
            self.setValue(self.singleStep() * steps)

    def decimals(self) -> int:
        return self._decimals

    def setDecimals(self, new_value: int) -> None:
        if new_value >= 0:
            self._decimals = new_value
        else:
            raise ValueError(f"Invalid decimals value: {new_value}")



#!/usr/bin/env python3
import atexit
import enum
import http
import shutil
import sys
from pathlib import Path
from tempfile import mkdtemp

if sys.version_info < (3, 8):
    message = (
        "The Python version "
        + ".".join(map(str, sys.version_info[:3]))
        + " is not supported.\n"
        + "Use Python 3.8 or newer."
    )
    try:
        import tkinter
    except ImportError:
        input(message)  # wait for the user to see the text
    else:
        print(message, file=sys.stderr)

        import tkinter.messagebox

        _root = tkinter.Tk()
        _root.withdraw()
        tkinter.messagebox.showerror(title="Outdated Python", message=message)
        _root.destroy()

    exit(1)

if sys.version_info < (3, 10):

    def list_files(path: Path, *, suffix: "str | None" = None) -> "list[Path]":
        files: "list[Path]" = []
        if path.name.startswith("."):
            # ignore hidden files
            return []
        if path.is_dir():
            for file in path.iterdir():
                files.extend(list_files(file, suffix=suffix))
        elif path.is_file() and (suffix in (None, path.suffix)):
            files.append(path.absolute())
        return files

    me: Path = Path(__file__).resolve()
    my_parent: Path = me.parent

    annotations_needed: bool = False
    for f in list_files(my_parent):
        if f.is_file():
            if f.suffix == me.suffix:
                lines: "list[str]" = f.read_text().splitlines()
                if not any(line.startswith("from __future__ import annotations") for line in lines):
                    annotations_needed = True

    if annotations_needed:
        tmp_dir: Path = Path(mkdtemp())
        sys.path.insert(0, str(tmp_dir))

        for f in list_files(my_parent):
            if f.is_file():
                (tmp_dir / f.relative_to(my_parent)).parent.mkdir(parents=True, exist_ok=True)
                if f.suffix == me.suffix:
                    lines: "list[str]" = f.read_text().splitlines()
                    if not any(line.startswith("from __future__ import annotations") for line in lines):
                        lines.insert(0, "from __future__ import annotations")
                    new_text: str = "\n".join(lines)
                    new_text = new_text.replace("ParamSpec", "TypeVar")
                    (tmp_dir / f.relative_to(my_parent)).write_text(new_text)
                else:
                    (tmp_dir / f.relative_to(my_parent)).write_bytes(f.read_bytes())
            elif f.is_dir():
                (tmp_dir / f.relative_to(my_parent)).mkdir()

        atexit.register(shutil.rmtree, tmp_dir, ignore_errors=True)

if sys.version_info < (3, 11):

    class HTTPMethod(enum.Enum):
        CONNECT = "CONNECT"
        DELETE = "DELETE"
        GET = "GET"
        HEAD = "HEAD"
        OPTIONS = "OPTIONS"
        PATCH = "PATCH"
        POST = "POST"
        PUT = "PUT"
        TRACE = "TRACE"

    http.HTTPMethod = HTTPMethod


if __name__ == "__main__":
    try:
        from pycatsearch import main
    except ImportError:
        __author__ = "StSav012"
        __original_name__ = "pycatsearch"

        try:
            from updater import update_with_pip

            update_with_pip(__original_name__)

            from pycatsearch import main_gui as main
        except ImportError:
            from updater import update_with_pip, update_from_github, update_with_git

            update_with_git() or update_from_github(__author__, __original_name__)

            from src.pycatsearch import main_gui as main
    main()



from contextlib import suppress
from typing import Any, Collection, TYPE_CHECKING

from qtpy.QtCore import QModelIndex, Qt, Signal, Slot
from qtpy.QtGui import QContextMenuEvent, QIcon
from qtpy.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from .html_style_delegate import HTMLDelegate
from .selectable_label import SelectableLabel
from .url_label import URLLabel
from ..catalog import Catalog
from ..utils import CatalogEntryType, HUMAN_READABLE, ID, INCHI_KEY, LINES, STATE_HTML, best_name, chem_html

__all__ = ["SubstanceInfoSelector", "SubstanceInfo"]


class SubstanceInfoSelector(QDialog):
    tagSelectionChanged: Signal = Signal(int, bool, name="tagSelectionChanged")

    def __init__(
        self,
        catalog: Catalog,
        species_tags: Collection[int],
        *,
        selected_species_tags: Collection[int] = (),
        allow_html: bool = True,
        inchi_key_search_url_template: str = "",
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)
        self._catalog: Catalog = catalog
        self._inchi_key_search_url_template: str = inchi_key_search_url_template
        self.setModal(True)
        self.setWindowTitle(self.tr("Select Substance"))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QVBoxLayout = QVBoxLayout(self)
        self._list_box: QListWidget = QListWidget(self)
        self._list_box.itemChanged.connect(self._on_list_item_changed)
        self._list_box.doubleClicked.connect(self._on_list_double_clicked)
        self._list_box.setItemDelegateForColumn(0, HTMLDelegate(self._list_box))
        layout.addWidget(self._list_box)
        species_tags = set(species_tags)
        while species_tags:
            species_tag: int = species_tags.pop()
            # don't specify the parent here: https://t.me/qtforpython/20950
            item: QListWidgetItem = QListWidgetItem(best_name(catalog.catalog[species_tag], allow_html=allow_html))
            item.setData(Qt.ItemDataRole.ToolTipRole, str(species_tag))
            item.setData(Qt.ItemDataRole.UserRole, species_tag)
            item.setCheckState(
                Qt.CheckState.Checked if species_tag in selected_species_tags else Qt.CheckState.Unchecked
            )
            self._list_box.addItem(item)
        self._buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        self._buttons.rejected.connect(self.reject)
        layout.addWidget(self._buttons)

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        context_menu: QMenu = QMenu(self)
        context_menu.addAction(
            self._icon(
                "dialog-information",
                "mdi6.flask-empty-outline",
                "mdi6.information-variant",
                options=[{}, {"scale_factor": 0.5}],
            ),
            self.tr("Substance &Info"),
            lambda: self._list_box.doubleClicked.emit(self._list_box.currentIndex()),
        )
        context_menu.exec(event.globalPos())
        return super().contextMenuEvent(event)

    def _icon(
        self,
        theme_name: str,
        *qta_name: str,
        standard_pixmap: QStyle.StandardPixmap | None = None,
        **qta_specs: Any,
    ) -> QIcon:
        if theme_name and QIcon.hasThemeIcon(theme_name):
            return QIcon.fromTheme(theme_name)

        if qta_name:
            with suppress(ImportError, Exception):
                import qtawesome as qta

                return qta.icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

        if standard_pixmap is not None:
            return self.style().standardIcon(standard_pixmap)

        return QIcon()

    @Slot(QListWidgetItem)
    def _on_list_item_changed(self, item: QListWidgetItem) -> None:
        self.tagSelectionChanged.emit(item.data(Qt.ItemDataRole.UserRole), item.checkState() == Qt.CheckState.Checked)

    @Slot(QModelIndex)
    def _on_list_double_clicked(self, index: QModelIndex) -> None:
        item: QListWidgetItem = self._list_box.item(index.row())
        syn: SubstanceInfo = SubstanceInfo(
            self._catalog,
            item.data(Qt.ItemDataRole.UserRole),
            inchi_key_search_url_template=self._inchi_key_search_url_template,
            parent=self,
        )
        syn.exec()


class SubstanceInfo(QDialog):
    """A simple dialog that displays the information about a substance from the loaded catalog"""

    def __init__(
        self,
        catalog: Catalog,
        species_tag: int,
        inchi_key_search_url_template: str = "",
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)
        self.setModal(True)
        self.setWindowTitle(self.tr("Substance Info"))
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())
        layout: QFormLayout = QFormLayout(self)
        label: SelectableLabel
        entry: CatalogEntryType = catalog.catalog[species_tag]
        for key in entry.__slots__:
            if key == LINES:
                continue
            elif key == ID:
                label = URLLabel(
                    url=f"https://cdms.astro.uni-koeln.de/cdms/portal/catalog/{getattr(entry, key)}/",
                    text=f"{getattr(entry, key)}",
                    parent=self,
                )
                label.setOpenExternalLinks(True)
            elif key == STATE_HTML:
                label = SelectableLabel(chem_html(str(getattr(entry, key))), self)
            elif key == INCHI_KEY and inchi_key_search_url_template:
                label = URLLabel(
                    url=inchi_key_search_url_template.format(InChIKey=getattr(entry, key)),
                    text=getattr(entry, key),
                    parent=self,
                )
                label.setOpenExternalLinks(True)
            else:
                label = SelectableLabel(str(getattr(entry, key)), self)
            layout.addRow(self.tr(HUMAN_READABLE[key]), label)
        label = SelectableLabel(str(len(entry.lines)), self)
        layout.addRow(self.tr("Number of spectral lines"), label)
        buttons: QDialogButtonBox = QDialogButtonBox(QDialogButtonBox.StandardButton.Close, self)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        # add the texts to the translation table but don't run the code at runtime
        if TYPE_CHECKING:
            self.tr("Catalog")
            self.tr("Lines")
            self.tr("Frequency")
            self.tr("Intensity")
            self.tr("ID")
            self.tr("Molecule")
            self.tr("Structural formula")
            self.tr("Stoichiometric formula")
            self.tr("Molecule symbol")
            self.tr("Species tag")
            self.tr("Name")
            self.tr("Trivial name")
            self.tr("Isotopolog")
            self.tr("State (TeX)")
            self.tr("State (HTML)")
            self.tr("InChI key")
            self.tr("Contributor")
            self.tr("Version")
            self.tr("Date of entry")
            self.tr("Degrees of freedom")
            self.tr("Lower state energy")



from math import nan
from os import PathLike
from pathlib import Path

from qtpy.QtWidgets import QDialog, QWidget

from .catalog_wizard import SaveCatalogWizard
from .catalog_wizard.download_confirmation_page import DownloadConfirmationPage
from .catalog_wizard.progress_page import ProgressPage
from .catalog_wizard.settings_page import SettingsPage
from .catalog_wizard.summary_page import SummaryPage
from .settings import Settings
from ..catalog import Catalog

__all__ = ["UpdateDialog"]


class UpdateDialog(SaveCatalogWizard):
    """GUI for `async_downloader.Downloader` or `downloader.Downloader`"""

    def __init__(
        self,
        settings: Settings,
        existing_catalog_location: str | PathLike[str],
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(settings=settings, parent=parent, default_save_location=Path(existing_catalog_location))

        self._old_catalog: Catalog = Catalog(existing_catalog_location)

        self.setWindowTitle(self.tr("Update Catalog"))

        self._settings_page: SettingsPage = SettingsPage(self)
        self.addPage(self._settings_page)
        self.addPage(DownloadConfirmationPage(self))
        self._progress_page: ProgressPage = ProgressPage(self, existing_catalog=self._old_catalog)
        self.addPage(self._progress_page)
        self.addPage(SummaryPage(self))

        if self._old_catalog:
            self._settings_page.frequency_limits = self.frequency_limits()
        else:
            self._progress_page.setCommitPage(True)
            self.setStartId(self._progress_page.nextId())

        # don't show `self._settings_page`
        self._settings_page.setCommitPage(True)
        self.setStartId(self._settings_page.nextId())

    def back(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().back()

    def next(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().next()

    def restart(self) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        super().restart()

    def frequency_limits(self) -> tuple[float, float]:
        if not self._old_catalog:
            return nan, nan
        return self._old_catalog.min_frequency, self._old_catalog.max_frequency

    def done(self, exit_code: QDialog.DialogCode) -> None:
        if self._progress_page.isActive():
            self._progress_page.stop()
        return super().done(exit_code)



import enum
import http
import platform
import sys
from argparse import ArgumentParser, Namespace, ZERO_OR_MORE
from importlib import import_module
from importlib.abc import Loader, MetaPathFinder
from importlib.machinery import ModuleSpec
from importlib.util import spec_from_file_location
from pathlib import Path
from types import ModuleType

__author__ = "StSav012"
__original_name__ = "pycatsearch"


try:
    from ._version import __version__
except ImportError:
    __version__ = ""

if sys.version_info < (3, 10) and __file__ != "<string>":

    class StringImporter(MetaPathFinder):
        class Loader(Loader):
            def __init__(self, modules: "dict[str, str | dict]") -> None:
                self._modules: "dict[str, str | dict]" = modules

            # noinspection PyMethodMayBeStatic
            def is_package(self, module_name: str) -> bool:
                return isinstance(self._modules[module_name], dict)

            # noinspection PyMethodMayBeStatic
            def get_code(self, module_name: str):
                return compile(self._modules[module_name], filename="<string>", mode="exec")

            def create_module(self, spec: ModuleSpec) -> "ModuleType | None":
                return ModuleType(spec.name)

            def exec_module(self, module: ModuleType) -> None:
                if module.__name__ not in self._modules:
                    raise ImportError(module.__name__)

                sys.modules[module.__name__] = module
                if not self.is_package(module.__name__):
                    exec(self._modules[module.__name__], module.__dict__)
                else:
                    for sub_module in self._modules[module.__name__]:
                        self._modules[".".join((module.__name__, sub_module))] = self._modules[module.__name__][
                            sub_module
                        ]
                    exec(self._modules[module.__name__].get("__init__", ""), module.__dict__)

        def __init__(self, **modules: "str | dict") -> None:
            self._modules: "dict[str, str | dict]" = modules
            self._loader = StringImporter.Loader(modules)

        def find_spec(
            self,
            fullname: str,
            path: "str | None",
            target: "ModuleType | None" = None,
        ) -> "ModuleSpec | None":
            if fullname in self._modules:
                spec: ModuleSpec = spec_from_file_location(fullname, loader=self._loader)
                spec.origin = "<string>"
                return spec
            return None

    def list_files(path: Path, *, suffix: "str | None" = None) -> "list[Path]":
        files: "list[Path]" = []
        if path.is_dir():
            for file in path.iterdir():
                files.extend(list_files(file, suffix=suffix))
        elif path.is_file() and (suffix in (None, path.suffix)):
            files.append(path.absolute())
        return files

    me: Path = Path(__file__).resolve()
    my_parent: Path = me.parent

    py38_modules: "dict[str, str | dict]" = {}

    for f in list_files(my_parent, suffix=me.suffix):
        lines: "list[str]" = f.read_text(encoding="utf-8").splitlines()
        if not any(line.startswith("from __future__ import annotations") for line in lines):
            lines.insert(0, "from __future__ import annotations")
            new_text: str = "\n".join(lines)
            new_text = new_text.replace("ParamSpec", "TypeVar")
            parts: "tuple[str, ...]" = f.relative_to(my_parent).parts
            p: "dict[str, str | dict]" = py38_modules
            for part in parts[:-1]:
                if part not in p:
                    p[part] = {}
                p = p[part]
            p[parts[-1][: -len(me.suffix)]] = new_text

    if py38_modules:
        for m in list(sys.modules):
            if m.startswith(__original_name__):
                if m in sys.modules:  # check again in case the module's gone midway
                    sys.modules.pop(m)

        sys.meta_path.insert(0, StringImporter(**{__original_name__: py38_modules}))
        if __original_name__ not in sys.modules:
            sys.modules[__original_name__] = import_module(__original_name__)

if sys.version_info < (3, 11):

    class HTTPMethod(enum.Enum):
        CONNECT = "CONNECT"
        DELETE = "DELETE"
        GET = "GET"
        HEAD = "HEAD"
        OPTIONS = "OPTIONS"
        PATCH = "PATCH"
        POST = "POST"
        PUT = "PUT"
        TRACE = "TRACE"

    http.HTTPMethod = HTTPMethod


def _argument_parser() -> ArgumentParser:
    ap: ArgumentParser = ArgumentParser(
        allow_abbrev=True,
        description="Yet another implementation of JPL and CDMS spectroscopy catalogs offline search.\n"
        f"Find more at https://github.com/{__author__}/{__original_name__}.",
    )
    ap.add_argument("catalog", type=Path, help="the catalog location to load", nargs=ZERO_OR_MORE)
    return ap


def _cli_argument_parser() -> ArgumentParser:
    ap: ArgumentParser = _argument_parser()
    ap.add_argument("-fmin", "--min-frequency", type=float, help="the lower frequency [MHz] to take")
    ap.add_argument("-fmax", "--max-frequency", type=float, help="the upper frequency [MHz] to take")
    ap.add_argument(
        "-imin",
        "--min-intensity",
        type=float,
        help="the minimal intensity [log10(nm²×MHz)] to take",
    )
    ap.add_argument(
        "-imax",
        "--max-intensity",
        type=float,
        help="the maximal intensity [log10(nm²×MHz)] to take",
    )
    ap.add_argument(
        "-T",
        "--temperature",
        type=float,
        help="the temperature [K] to calculate the line intensity at, use the catalog intensity if not set",
    )
    ap.add_argument(
        "-t",
        "--tag",
        "--species-tag",
        type=int,
        dest="species_tag",
        help="a number to match the `speciestag` field",
    )
    ap.add_argument(
        "-n",
        "--any-name-or-formula",
        type=str,
        help="a string to match any field used by `any_name` and `any_formula` options",
    )
    ap.add_argument("-a", "--anything", type=str, help="a string to match any field")
    ap.add_argument("--any-name", type=str, help="a string to match the `trivial name` or the `name` field")
    ap.add_argument(
        "--any-formula",
        type=str,
        help="a string to match the `structuralformula`, `moleculesymbol`, `stoichiometricformula`, or `isotopolog` field",
    )
    ap.add_argument(
        "--InChI-key",
        "--inchi-key",
        type=str,
        dest="inchi_key",
        help="a string to match the `inchikey` field, which contains the IUPAC International Chemical Identifier (InChI™)",
    )
    ap.add_argument("--trivial-name", type=str, help="a string to match the `trivial name` field")
    ap.add_argument("--structural-formula", type=str, help="a string to match the `structural formula` field")
    ap.add_argument("--name", type=str, help="a string to match the `name` field")
    ap.add_argument("--stoichiometric-formula", type=str, help="a string to match the `stoichiometric formula` field")
    ap.add_argument("--isotopolog", type=str, help="a string to match the `isotopolog` field")
    ap.add_argument("--state", type=str, help="a string to match the `state` or `state_html` field")
    ap.add_argument(
        "--dof",
        "--degrees_of_freedom",
        type=int,
        dest="degrees_of_freedom",
        help="0 for atoms, 2 for linear molecules, and 3 for nonlinear molecules",
    )

    return ap


def main_cli() -> int:
    ap: ArgumentParser = _cli_argument_parser()
    args: Namespace = ap.parse_intermixed_args()

    search_args: dict[str, str | float | int] = dict(
        (key, value) for key, value in args.__dict__.items() if key != "catalog" and value is not None
    )
    if any(value is not None for value in search_args.values()):
        from .catalog import Catalog

        c: Catalog = Catalog(*args.catalog)
        c.print(**search_args)
        return 0
    else:
        print("No search parameter specified", file=sys.stderr)
        ap.print_help(file=sys.stderr)
        return 1


def _show_exception(ex: Exception) -> None:
    from traceback import format_exception

    error_message: str = ""
    if isinstance(ex, SyntaxError):
        error_message = "Python %s is not supported.\nGet a newer Python!" % platform.python_version()
    elif isinstance(ex, ImportError):
        if ex.name is not None:
            if "from" in ex.msg.split():
                error_message = (
                    "Module %s lacks a part, or the latter cannot be loaded for a reason.\n"
                    "Try to update the module." % repr(ex.name)
                )
            elif ex.path is None:
                error_message = "Module %s cannot be found.\nTry to install it." % repr(ex.name)
            else:
                error_message = (
                    "Module %s cannot be loaded for an unspecified reason.\n"
                    "Try to install or reinstall it." % repr(ex.name)
                )
        else:
            error_message = str(ex)
    if error_message:
        error_message += "\n"

    error_message += "".join(format_exception(*sys.exc_info()))

    print(error_message, file=sys.stderr)

    try:
        import tkinter
        import tkinter.messagebox
    except (ModuleNotFoundError, ImportError):
        pass
    else:
        root: tkinter.Tk = tkinter.Tk()
        root.withdraw()
        if isinstance(ex, SyntaxError):
            tkinter.messagebox.showerror(title="Syntax Error", message=error_message)
        elif isinstance(ex, ImportError):
            tkinter.messagebox.showerror(title="Package Missing", message=error_message)
        else:
            tkinter.messagebox.showerror(title="Error", message=error_message)
        root.destroy()


def main_gui() -> int:
    ap: ArgumentParser = _argument_parser()
    args: Namespace = ap.parse_intermixed_args()

    try:
        from . import gui
    except Exception as ex:
        _show_exception(ex)
        return 1
    else:
        try:
            return gui.run(*args.catalog)
        except Exception as ex:
            _show_exception(ex)
            return 1


def download() -> None:
    from . import downloader

    downloader.download()


def async_download() -> None:
    from . import async_downloader

    async_downloader.download()



#!/usr/bin/env python3


def _third_party_modules() -> list[str]:
    import site
    import sys

    prefixes: list[str] = site.getsitepackages([sys.exec_prefix, sys.prefix])
    third_party_modules: list[str] = []
    for module_name, module in sys.modules.copy().items():
        paths = getattr(module, "__path__", [])
        if (
            "." not in module_name
            and module_name != "_distutils_hack"
            and paths
            and getattr(module, "__package__", "")
            and any(p.startswith(prefix) for p in paths for prefix in prefixes)
        ):
            third_party_modules.append(module_name)

    return third_party_modules


def test_gui():
    from pycatsearch import main_gui as main

    third_party_modules: list[str]

    third_party_modules = _third_party_modules()
    assert third_party_modules == [], third_party_modules

    # intentionally fail the app after letting it import everything it needs
    from qtpy.QtCore import QCoreApplication

    app = QCoreApplication()
    assert (r := main()) != 0, r
    assert app is not None  # to ensure the app still exists

    expected_third_party_modules: list[str] = [
        "PyQt5",
        "PyQt6",
        "PySide2",
        "PySide6",
        "aiohttp",
        "aiosignal",
        "attr",
        "frozenlist",
        "idna",
        "multidict",
        "orjson",
        "packaging",
        "qtawesome",
        "qtpy",
        "shiboken2",
        "shiboken6",
        "yarl",
    ]
    third_party_modules = _third_party_modules()
    assert third_party_modules
    assert set(third_party_modules).issubset(expected_third_party_modules), third_party_modules


if __name__ == "__main__":
    import sys
    from os import path

    sys.path = list(set(sys.path) | {path.abspath(path.join(__file__, path.pardir, path.pardir))})

    test_gui()



import sys
from contextlib import suppress
from datetime import datetime, timedelta, timezone
from functools import partialmethod
from os import PathLike
from pathlib import Path
from typing import Any

from qtpy import PYSIDE2, QT6
from qtpy.QtCore import QLibraryInfo, QLocale, QTranslator, Qt, qVersion
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QAbstractSpinBox, QApplication, QDialog, QMenu

__all__ = ["qta_icon", "run"]


def _version_tuple(version_string: bytes | str) -> tuple[int | bytes | str, ...]:
    result: tuple[int | bytes | str, ...] = tuple()
    part: bytes | str
    for part in version_string.split("." if isinstance(version_string, str) else b"."):
        try:
            result += (int(part),)
        except ValueError:
            # follow `pkg_resources` version 0.6a9: remove dashes to sort letters after digits
            result += (part.replace("-", ""),) if isinstance(part, str) else (part.replace(b"-", b""),)
    return result


def _warn_about_outdated_package(package_name: str, package_version: str, release_time: datetime) -> None:
    """Display a warning about an outdated package a year after the package released"""
    if datetime.now(tz=timezone.utc) - release_time > timedelta(days=366):
        import tkinter.messagebox

        tkinter.messagebox.showwarning(
            title="Package Outdated", message=f"Please update {package_name} package to {package_version} or newer"
        )


def _make_old_qt_compatible_again() -> None:
    # noinspection PyUnresolvedReferences
    def to_iso_format(s: str) -> str:
        if sys.version_info < (3, 11):
            import re
            from typing import Callable

            if s.endswith("Z"):
                # '2011-11-04T00:05:23Z'
                s = s[:-1] + "+00:00"

            def from_iso_datetime(m: re.Match[str]) -> str:
                groups: dict[str, str] = m.groupdict("")
                date: str = f"{m['year']}-{m['month']}-{m['day']}"
                time: str = (
                    f"{groups['hour']:0>2}:{groups['minute']:0>2}:{groups['second']:0>2}.{groups['fraction']:0<6}"
                )
                return date + "T" + time + groups["offset"]

            # noinspection PyUnresolvedReferences
            def from_iso_calendar(m: re.Match[str]) -> str:
                from datetime import date

                groups: dict[str, str] = m.groupdict("")
                date: str = date.fromisocalendar(
                    year=int(m["year"]), week=int(m["week"]), day=int(m["dof"])
                ).isoformat()
                time: str = (
                    f"{groups['hour']:0>2}:{groups['minute']:0>2}:{groups['second']:0>2}.{groups['fraction']:0<6}"
                )
                return date + "T" + time + groups["offset"]

            patterns: dict[str, Callable[[re.Match[str]], str]] = {
                # '20111104', '20111104T000523283'
                r"(?P<year>\d{4})(?P<month>\d{2})(?P<day>\d{2})"
                r"(.(?P<hour>\d{2})(?P<minute>\d{2})(?P<second>\d{2})(?P<fraction>\d+)?)?"
                r"(?P<offset>[+\-].+)?": from_iso_datetime,
                # '2011-11-04', '2011-11-04T00:05:23.283', '2011-11-04T00:05:23.283+00:00'
                r"(?P<year>\d{4})-(?P<month>\d{1,2})-(?P<day>\d{1,2})"
                r"(.(?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})(\.(?P<fraction>\d+))?)?"
                r"(?P<offset>[+\-].+)?": from_iso_datetime,
                # '2011-W01-2T00:05:23.283'
                r"(?P<year>\d{4})-W(?P<week>\d{1,2})-(?P<dof>\d{1,2})"
                r"(.(?P<hour>\d{1,2}):(?P<minute>\d{1,2}):(?P<second>\d{1,2})(\.(?P<fraction>\d+))?)?"
                r"(?P<offset>[+\-].+)?": from_iso_calendar,
                # '2011W0102T000523283'
                r"(?P<year>\d{4})-W(?P<week>\d{2})-(?P<dof>\d{2})"
                r"(.(?P<hour>\d{1,2})(?P<minute>\d{1,2})(?P<second>\d{1,2})(?P<fraction>\d+)?)?"
                r"(?P<offset>[+\-].+)?": from_iso_calendar,
            }
            match: re.Match[str] | None
            for p in patterns:
                match = re.fullmatch(p, s)
                if match is not None:
                    s = patterns[p](match)
                    break

        return s

    if not QT6:
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_EnableHighDpiScaling)
        QApplication.setAttribute(Qt.ApplicationAttribute.AA_UseHighDpiPixmaps)

    from qtpy import __version__

    if _version_tuple(__version__) < _version_tuple("2.3.1"):
        _warn_about_outdated_package(
            package_name="QtPy",
            package_version="2.3.1",
            release_time=datetime.fromisoformat(to_iso_format("2023-03-28T23:06:05Z")),
        )
        if QT6:
            QLibraryInfo.LibraryLocation = QLibraryInfo.LibraryPath
    if _version_tuple(__version__) < _version_tuple("2.4.0"):
        _warn_about_outdated_package(
            package_name="QtPy",
            package_version="2.4.0",
            release_time=datetime.fromisoformat(to_iso_format("2023-08-29T16:24:56Z")),
        )
        if PYSIDE2:
            QApplication.exec = QApplication.exec_
            QDialog.exec = QDialog.exec_
            QMenu.exec = lambda self, pos: self.exec_(pos)

        if not QT6:
            QLibraryInfo.path = lambda *args, **kwargs: QLibraryInfo.location(*args, **kwargs)
            QLibraryInfo.LibraryPath = QLibraryInfo.LibraryLocation

        if _version_tuple(qVersion()) < _version_tuple("6.3"):
            from qtpy.QtCore import QObject
            from qtpy.QtGui import QKeySequence
            from qtpy.QtWidgets import QAction, QToolBar, QWidget

            def add_action(self: QWidget, *args, old_add_action) -> QAction:
                action: QAction
                icon: QIcon
                text: str
                shortcut: QKeySequence | QKeySequence.StandardKey | str | int
                receiver: QObject
                member: bytes
                if all(
                    isinstance(arg, t)
                    for arg, t in zip(args, [str, (QKeySequence, QKeySequence.StandardKey, str, int), QObject, bytes])
                ):
                    if len(args) == 2:
                        text, shortcut = args
                        action = old_add_action(self, text)
                        action.setShortcut(shortcut)
                    elif len(args) == 3:
                        text, shortcut, receiver = args
                        action = old_add_action(self, text, receiver)
                        action.setShortcut(shortcut)
                    elif len(args) == 4:
                        text, shortcut, receiver, member = args
                        action = old_add_action(self, text, receiver, member, shortcut)
                    else:
                        return old_add_action(self, *args)
                    return action
                elif all(
                    isinstance(arg, t)
                    for arg, t in zip(
                        args, [QIcon, str, (QKeySequence, QKeySequence.StandardKey, str, int), QObject, bytes]
                    )
                ):
                    if len(args) == 3:
                        icon, text, shortcut = args
                        action = old_add_action(self, icon, text)
                        action.setShortcut(QKeySequence(shortcut))
                    elif len(args) == 4:
                        icon, text, shortcut, receiver = args
                        action = old_add_action(self, icon, text, receiver)
                        action.setShortcut(QKeySequence(shortcut))
                    elif len(args) == 5:
                        icon, text, shortcut, receiver, member = args
                        action = old_add_action(self, icon, text, receiver, member, QKeySequence(shortcut))
                    else:
                        return old_add_action(self, *args)
                    return action
                return old_add_action(self, *args)

            QMenu.addAction = partialmethod(add_action, old_add_action=QMenu.addAction)
            QToolBar.addAction = partialmethod(add_action, old_add_action=QToolBar.addAction)
    if _version_tuple(__version__) < _version_tuple("2.4.1"):
        _warn_about_outdated_package(
            package_name="QtPy",
            package_version="2.4.1",
            release_time=datetime.fromisoformat(to_iso_format("2023-10-23T23:57:23Z")),
        )

    # not a part of any QtPy (yet)
    if PYSIDE2:
        # noinspection PyUnresolvedReferences
        QAbstractSpinBox.setAlignment = partialmethod(
            lambda self, flag, _old: _old(self, Qt.Alignment(flag)),
            _old=QAbstractSpinBox.setAlignment,
        )
        Qt.AlignmentFlag.__or__ = lambda self, other: int(self) | int(other)
        Qt.AlignmentFlag.__ror__ = lambda self, other: int(other) | int(self)


def qta_icon(*qta_name: str, **qta_specs: Any) -> QIcon:
    if qta_name:
        with suppress(ImportError, Exception):
            from qtawesome import icon

            return icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

    return QIcon()


def run(*catalogs: str | PathLike[str]) -> int:
    import re

    # fix `re.RegexFlag.NOFLAG` missing on some systems
    if not hasattr(re.RegexFlag, "NOFLAG"):
        re.RegexFlag.NOFLAG = 0

    _make_old_qt_compatible_again()

    from .ui import UI
    from ..catalog import Catalog

    app: QApplication = QApplication(sys.argv)

    languages: set[str] = set(QLocale().uiLanguages() + [QLocale().bcp47Name(), QLocale().name()])
    language: str
    translations_path: str = QLibraryInfo.path(QLibraryInfo.LibraryPath.TranslationsPath)
    qt_translator: QTranslator = QTranslator()
    for language in languages:
        if qt_translator.load("qt_" + language, translations_path):
            QApplication.installTranslator(qt_translator)
            break
    qtbase_translator: QTranslator = QTranslator()
    for language in languages:
        if qtbase_translator.load("qtbase_" + language, translations_path):
            QApplication.installTranslator(qtbase_translator)
            break
    my_translator: QTranslator = QTranslator()
    for language in languages:
        if my_translator.load(language, str(Path(__file__).parent / "i18n")):
            QApplication.installTranslator(my_translator)
            break

    window: UI = UI(Catalog(*catalogs))
    window.show()
    return app.exec()



import sys
from math import inf
from pathlib import Path
from typing import Any, final

from qtpy.QtCore import QItemSelection, QMimeData, QModelIndex, QPoint, Qt, Slot
from qtpy.QtGui import QClipboard, QCloseEvent, QCursor, QIcon, QPalette, QPixmap, QScreen
from qtpy.QtWidgets import (
    QAbstractItemView,
    QAbstractSpinBox,
    QApplication,
    QDoubleSpinBox,
    QFormLayout,
    QHeaderView,
    QMainWindow,
    QMessageBox,
    QPushButton,
    QSizePolicy,
    QSplitter,
    QStatusBar,
    QTableView,
    QVBoxLayout,
    QWidget,
)

from .about_dialog import about
from .catalog_file_dialog import CatalogOpenFileDialog, CatalogSaveFileDialog
from .catalog_info import CatalogInfo
from .download_dialog import DownloadDialog
from .float_spinbox import FloatSpinBox
from .found_lines_model import FoundLinesModel
from .frequency_box import FrequencyBox
from .html_style_delegate import HTMLDelegate
from .menu_bar import MenuBar
from .preferences import Preferences
from .save_catalog_waiting_screen import SaveCatalogWaitingScreen
from .settings import Settings
from .substance_info import SubstanceInfo
from .substances_box import SubstanceBox
from .. import __version__
from ..catalog import Catalog, CatalogType
from ..utils import ReleaseInfo, a_tag, latest_release, p_tag, remove_html, tag, update_with_pip, wrap_in_html

if sys.version_info < (3, 10):
    # noinspection PyUnresolvedReferences
    from ..utils import zip

__all__ = ["UI"]


def copy_to_clipboard(text: str, text_type: Qt.TextFormat | str = Qt.TextFormat.PlainText) -> None:
    clipboard: QClipboard = QApplication.clipboard()
    if not text:
        return
    mime_data: QMimeData = QMimeData()
    if isinstance(text_type, str):
        mime_data.setData(text_type, text.encode())
    elif text_type == Qt.TextFormat.RichText:
        mime_data.setHtml(wrap_in_html(text))
        mime_data.setText(remove_html(text))
    else:
        mime_data.setText(text)
    clipboard.setMimeData(mime_data, QClipboard.Mode.Clipboard)


@final
class UI(QMainWindow):
    def __init__(self, catalog: Catalog, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setObjectName("mainWindow")

        self.catalog: Catalog = catalog
        self.settings: Settings = Settings("SavSoft", "CatSearch", self)

        self.open_dialog: CatalogOpenFileDialog = CatalogOpenFileDialog(settings=self.settings, parent=self)
        self.save_dialog: CatalogSaveFileDialog = CatalogSaveFileDialog(settings=self.settings, parent=self)

        self._central_widget: QSplitter = QSplitter(Qt.Orientation.Vertical, self)
        self._central_widget.setObjectName("horizontalSplitter")
        self._top_matter: QSplitter = QSplitter(Qt.Orientation.Horizontal, self._central_widget)
        self._top_matter.setObjectName("verticalSplitter")
        self._right_matter: QWidget = QWidget(self._central_widget)

        self.spin_intensity: FloatSpinBox = FloatSpinBox(self._central_widget)
        self.spin_temperature: QDoubleSpinBox = QDoubleSpinBox(self._central_widget)

        self.box_substance: SubstanceBox = SubstanceBox(self.catalog, self.settings, self._central_widget)
        self.box_frequency: FrequencyBox = FrequencyBox(self.settings, self._central_widget)
        self.button_search: QPushButton = QPushButton(self._central_widget)

        self.results_model: FoundLinesModel = FoundLinesModel(self.settings, self)
        self.results_table: QTableView = QTableView(self._central_widget)

        self.menu_bar: MenuBar = MenuBar(self)

        self.status_bar: QStatusBar = QStatusBar(self)

        def setup_ui() -> None:
            from . import qta_icon  # import locally to avoid a circular import

            def icon_from_data(data: bytes) -> QIcon:
                # https://ru.stackoverflow.com/a/1032610
                palette: QPalette = self.palette()
                pixmap: QPixmap = QPixmap()
                pixmap.loadFromData(
                    data.strip()
                    .replace(b'"background"', b'"' + palette.window().color().name().encode() + b'"')
                    .replace(b'"foreground"', b'"' + palette.text().color().name().encode() + b'"')
                )
                return QIcon(pixmap)

            # language=SVG
            window_icon: bytes = b"""\
            <svg height="64" width="64" version="1.1">
            <path stroke-linejoin="round" d="m6.722 8.432c-9.05 9.648-6.022 27.23 6.048 33.04 6.269 3.614 13.88 \
            3.1 20-0.1664l20 20c2.013 2.013 5.256 2.013 7.27 0l1.259-1.259c2.013-2.013 2.013-5.256 \
            0-7.27l-19.83-19.83c1.094-1.948 1.868-4.095 2.211-6.403 3.06-13.5-9.72-27.22-23.4-25.12-4.74 \
            0.53-9.28 2.72-12.64 6.104-0.321 0.294-0.626 0.597-0.918 0.908zm8.015 6.192c4.978-5.372 14.79-3.878 17.96 \
            2.714 3.655 6.341-0.6611 15.28-7.902 16.36-7.14 1.62-14.4-5.14-13.29-12.38 0.2822-2.51 1.441-4.907 \
            3.231-6.689z" stroke="background" stroke-width="2.4" fill="foreground"/>
            </svg>"""
            self.setWindowIcon(icon_from_data(window_icon))

            if __version__:
                self.setWindowTitle(self.tr("PyCatSearch (version {0})").format(__version__))
            else:
                self.setWindowTitle(self.tr("PyCatSearch"))
            self.setCentralWidget(self._central_widget)

            layout_right: QVBoxLayout = QVBoxLayout()
            layout_options: QFormLayout = QFormLayout()

            self.results_table.setModel(self.results_model)
            self.results_table.setItemDelegateForColumn(0, HTMLDelegate(self.results_table))
            self.results_table.setMouseTracking(True)
            self.results_table.setContextMenuPolicy(Qt.ContextMenuPolicy.CustomContextMenu)
            self.results_table.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
            self.results_table.setDropIndicatorShown(False)
            self.results_table.setDragDropOverwriteMode(False)
            self.results_table.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
            self.results_table.setCornerButtonEnabled(False)
            self.results_table.setSortingEnabled(True)
            self.results_table.setAlternatingRowColors(True)
            self.results_table.horizontalHeader().setObjectName("resultsTableHorizontalHeader")
            self.results_table.horizontalHeader().setDefaultSectionSize(180)
            self.results_table.horizontalHeader().setHighlightSections(False)
            self.results_table.horizontalHeader().setSectionResizeMode(0, QHeaderView.ResizeMode.Stretch)
            self.results_table.horizontalHeader().setSectionResizeMode(1, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionResizeMode(2, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionResizeMode(3, QHeaderView.ResizeMode.ResizeToContents)
            self.results_table.horizontalHeader().setSectionsMovable(True)
            self.results_table.verticalHeader().setVisible(False)
            self.results_table.verticalHeader().setHighlightSections(False)

            # substance selection
            self._top_matter.addWidget(self.box_substance)

            # frequency limits
            layout_right.addWidget(self.box_frequency, 1)

            self.spin_intensity.setAlignment(
                Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignTrailing | Qt.AlignmentFlag.AlignVCenter
            )
            self.spin_intensity.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
            self.spin_intensity.setDecimals(2)
            self.spin_intensity.setRange(-inf, inf)
            self.spin_intensity.setSingleStep(0.1)
            self.spin_intensity.setValue(-6.54)
            self.spin_intensity.setSizePolicy(
                QSizePolicy.Policy.Expanding, self.spin_intensity.sizePolicy().verticalPolicy()
            )
            self.spin_intensity.setStatusTip(self.tr("Limit shown spectral lines"))
            layout_options.addRow(self.tr("Minimal Intensity:"), self.spin_intensity)
            self.spin_temperature.setAlignment(
                Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignTrailing | Qt.AlignmentFlag.AlignVCenter
            )
            self.spin_temperature.setButtonSymbols(QAbstractSpinBox.ButtonSymbols.NoButtons)
            self.spin_temperature.setMaximum(999.99)
            self.spin_temperature.setValue(300.0)
            self.spin_temperature.setSizePolicy(
                QSizePolicy.Policy.Expanding, self.spin_temperature.sizePolicy().verticalPolicy()
            )
            self.spin_temperature.setStatusTip(self.tr("Temperature to calculate intensity"))
            self.spin_temperature.setSuffix(self.tr(" K"))
            layout_options.addRow(self.tr("Temperature:"), self.spin_temperature)
            layout_right.addLayout(layout_options, 0)

            self.button_search.setText(self.tr("Show"))
            layout_right.addWidget(self.button_search, 0)

            self._right_matter.setLayout(layout_right)
            self._top_matter.addWidget(self._right_matter)
            self._top_matter.setStretchFactor(0, 1)
            self._top_matter.setChildrenCollapsible(False)

            self._central_widget.addWidget(self._top_matter)
            self._central_widget.addWidget(self.results_table)
            self._central_widget.setStretchFactor(1, 1)
            self._central_widget.setChildrenCollapsible(False)

            self.setMenuBar(self.menu_bar)
            self.setStatusBar(self.status_bar)

            self.button_search.setShortcut("Ctrl+Return")

            self.button_search.setIcon(qta_icon("mdi6.magnify"))

            self.adjustSize()

        setup_ui()

        self.temperature: float = 300.0  # [K]
        self.minimal_intensity: float = -inf  # [log10(nm²×MHz)]

        self.button_search.setDisabled(self.catalog.is_empty)

        self.preferences_dialog: Preferences = Preferences(self.settings, self)

        self.preset_table()

        self.load_settings()

        self.results_table.customContextMenuRequested.connect(self._on_table_context_menu_requested)
        self.results_table.selectionModel().selectionChanged.connect(self._on_table_item_selection_changed)
        self.results_table.doubleClicked.connect(self._on_action_substance_info_triggered)
        self.spin_intensity.valueChanged.connect(self._on_spin_intensity_changed)
        self.spin_temperature.valueChanged.connect(self._on_spin_temperature_changed)
        self.button_search.clicked.connect(self._on_search_requested)
        self.box_frequency.frequencyLimitsChanged.connect(self._on_search_requested)
        self.box_substance.selectedSubstancesChanged.connect(self._on_search_requested)
        self.menu_bar.action_load.triggered.connect(self._on_action_load_triggered)
        self.menu_bar.action_reload.triggered.connect(self._on_action_reload_triggered)
        self.menu_bar.action_save_as.triggered.connect(self._on_action_save_as_triggered)
        self.menu_bar.action_download_catalog.triggered.connect(self._on_action_download_catalog_triggered)
        self.menu_bar.action_preferences.triggered.connect(self._on_action_preferences_triggered)
        self.menu_bar.action_quit.triggered.connect(self._on_action_quit_triggered)
        self.menu_bar.action_check_updates.triggered.connect(self._on_action_check_updates_triggered)
        self.menu_bar.action_about_catalogs.triggered.connect(self._on_action_about_catalogs_triggered)
        self.menu_bar.action_about.triggered.connect(self._on_action_about_triggered)
        self.menu_bar.action_about_qt.triggered.connect(self._on_action_about_qt_triggered)
        self.menu_bar.action_copy.triggered.connect(self._on_action_copy_triggered)
        self.menu_bar.action_select_all.triggered.connect(self._on_action_select_all_triggered)
        self.menu_bar.action_copy_current.triggered.connect(self._on_action_copy_current_triggered)
        self.menu_bar.action_copy_name.triggered.connect(self._on_action_copy_name_triggered)
        self.menu_bar.action_copy_frequency.triggered.connect(self._on_action_copy_frequency_triggered)
        self.menu_bar.action_copy_intensity.triggered.connect(self._on_action_copy_intensity_triggered)
        self.menu_bar.action_copy_lower_state_energy.triggered.connect(
            self._on_action_copy_lower_state_energy_triggered
        )
        self.menu_bar.action_show_substance.toggled.connect(self._on_action_show_substance_toggled)
        self.menu_bar.action_show_frequency.toggled.connect(self._on_action_show_frequency_toggled)
        self.menu_bar.action_show_intensity.toggled.connect(self._on_action_show_intensity_toggled)
        self.menu_bar.action_show_lower_state_energy.toggled.connect(self._on_action_show_lower_state_energy_toggled)
        self.menu_bar.action_substance_info.triggered.connect(self._on_action_substance_info_triggered)
        self.menu_bar.action_clear.triggered.connect(self._on_action_clear_triggered)

        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)

        if self.settings.check_updates:
            _latest_release: ReleaseInfo = latest_release()
            if (
                _latest_release
                and _latest_release.version != self.settings.ignored_version
                and _latest_release.version > __version__
            ):
                res: QMessageBox.StandardButton = QMessageBox.question(
                    self,
                    self.tr("Release Info"),
                    self.tr(
                        "Version {release.version} published {release.pub_date} is available. "
                        "Would you like to get the update? "
                        "The app will try to restart."
                    ).format(release=_latest_release),
                    QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Ignore,
                )
                if res == QMessageBox.StandardButton.Yes:
                    update_with_pip()
                elif res == QMessageBox.StandardButton.Ignore:
                    self.settings.ignored_version = _latest_release.version

    def closeEvent(self, event: QCloseEvent) -> None:
        self.save_settings()
        event.accept()

    def load_catalog(self, *catalog_file_names: Path) -> bool:
        self.setDisabled(True)
        last_cursor: QCursor = self.cursor()
        self.setCursor(Qt.CursorShape.WaitCursor)
        self.repaint()
        self.catalog = Catalog(*catalog_file_names)
        self.box_substance.catalog = self.catalog
        self.setCursor(last_cursor)
        self.setEnabled(True)
        self.button_search.setDisabled(self.catalog.is_empty)
        self.menu_bar.action_save_as.setDisabled(self.catalog.is_empty)
        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)
        return not self.catalog.is_empty

    @Slot(float)
    def _on_spin_temperature_changed(self, arg1: float) -> None:
        self.temperature = self.settings.to_k(arg1)
        self.fill_table()

    @Slot(float)
    def _on_spin_intensity_changed(self, arg1: float) -> None:
        self.minimal_intensity = self.settings.to_log10_sq_nm_mhz(arg1)
        self.fill_table()

    @Slot(QPoint)
    def _on_table_context_menu_requested(self, pos: QPoint) -> None:
        self.menu_bar.menu_edit.popup(self.results_table.viewport().mapToGlobal(pos))

    @Slot(QItemSelection, QItemSelection)
    def _on_table_item_selection_changed(self, _selected: QItemSelection, _deselected: QItemSelection) -> None:
        self.menu_bar.action_copy.setEnabled(self.results_table.selectionModel().hasSelection())
        self.menu_bar.action_substance_info.setEnabled(self.results_table.selectionModel().hasSelection())

    @Slot()
    def _on_action_load_triggered(self) -> None:
        self.status_bar.showMessage(self.tr("Select a catalog file to load."))
        new_catalog_filenames: list[str] = self.open_dialog.get_open_filenames()

        if new_catalog_filenames:
            self.status_bar.showMessage(self.tr("Loading…"))
            if self.load_catalog(*new_catalog_filenames):
                self.status_bar.showMessage(self.tr("Catalogs loaded."))
            else:
                self.status_bar.showMessage(self.tr("Failed to load a catalog."))

        else:
            self.status_bar.clearMessage()

    @Slot()
    def _on_action_reload_triggered(self) -> None:
        if self.catalog.sources:
            self.status_bar.showMessage(self.tr("Loading…"))
            if self.load_catalog(*self.catalog.sources):
                self.status_bar.showMessage(self.tr("Catalogs loaded."))
            else:
                self.status_bar.showMessage(self.tr("Failed to load a catalog."))
        else:
            self.status_bar.clearMessage()

    @Slot()
    def _on_action_save_as_triggered(self) -> None:
        catalog: CatalogType = self.catalog.catalog
        frequency_limits: tuple[float, float] = (self.catalog.min_frequency, self.catalog.max_frequency)
        save_filename: Path | None
        while True:
            if not (save_filename := self.save_dialog.get_save_filename()):
                return

            try:
                ws = SaveCatalogWaitingScreen(
                    self,
                    filename=save_filename,
                    catalog=catalog,
                    frequency_limits=frequency_limits,
                )
                ws.exec()
            except OSError as ex:
                QMessageBox.warning(
                    self,
                    self.tr("Unable to save the catalog"),
                    self.tr("Error {exception} occurred while saving {filename}. Try another location.").format(
                        exception=ex,
                        filename=save_filename,
                    ),
                )
            else:
                return

    def stringify_selection_html(self) -> str:
        """
        Convert selected rows to string for copying as rich text
        :return: the rich text representation of the selected table lines
        """
        if not self.results_table.selectionModel().hasSelection():
            return ""

        units: list[str] = [
            "",
            self.settings.frequency_unit_str,
            self.settings.intensity_unit_str,
            self.settings.energy_unit_str,
        ]
        with_units: bool = self.settings.with_units
        csv_separator: str = self.settings.csv_separator
        actions_checked: list[bool] = [_a.isChecked() for _a in self.menu_bar.menu_columns.actions()]

        def format_value(value: Any, unit: str) -> str:
            return (
                self.tr("{value} {unit}", "format value in html").format(value=value, unit=unit)
                if with_units and unit
                else self.tr("{value}", "format value in html").format(value=value)
            )

        columns_order: list[int] = [
            self.results_table.horizontalHeader().logicalIndex(_c)
            for _c, _a in zip(range(self.results_table.horizontalHeader().count()), actions_checked, strict=True)
            if _a
        ]
        text: list[str] = ["<table>"]
        values: list[str]
        index: QModelIndex
        for index in self.results_table.selectionModel().selectedRows():
            row: FoundLinesModel.DataType = self.results_model.row(index.row())
            values = [
                format_value(_v, _u)
                for _u, _v, _a in zip(
                    units,
                    (row.name, row.frequency_str, row.intensity_str, row.lower_state_energy_str),
                    actions_checked,
                    strict=True,
                )
                if _a
            ]
            text.append(
                "<tr><td>" + f"</td>{csv_separator}<td>".join(values[_c] for _c in columns_order) + "</td></tr>"
            )
        text.append("</table>")
        return self.settings.line_end.join(text)

    @Slot()
    def _on_action_download_catalog_triggered(self) -> None:
        downloader: DownloadDialog = DownloadDialog(
            settings=self.settings,
            frequency_limits=(self.catalog.min_frequency, self.catalog.max_frequency),
            parent=self,
        )
        downloader.exec()

    @Slot()
    def _on_action_preferences_triggered(self) -> None:
        if self.preferences_dialog.exec() == Preferences.DialogCode.Accepted:
            self.box_frequency.blockSignals(True)
            self.spin_intensity.blockSignals(True)
            self.spin_temperature.blockSignals(True)
            try:
                self.fill_parameters()
            except LookupError:
                raise
            finally:
                self.box_frequency.blockSignals(False)
                self.spin_intensity.blockSignals(False)
                self.spin_temperature.blockSignals(False)

            if self.results_model.rowCount():
                self.fill_table()
            else:
                self.preset_table()

    @Slot()
    def _on_action_quit_triggered(self) -> None:
        self.close()

    @Slot()
    def _on_action_clear_triggered(self) -> None:
        self.results_model.clear()
        self.preset_table()

    def copy_selected_items(self, col: int) -> None:
        if col >= self.results_model.columnCount():
            return

        def html_list(lines: list[str]) -> str:
            return "<ul><li>" + f"</li>{self.settings.line_end}<li>".join(lines) + "</li></ul>"

        text_to_copy: list[str] = []
        index: QModelIndex
        for index in self.results_table.selectionModel().selectedRows(col) or [
            self.results_table.selectionModel().currentIndex()
        ]:
            if index.isValid():
                text_to_copy.append(self.results_model.data(index))
        if not text_to_copy:
            return
        if col == 0:
            copy_to_clipboard(html_list(text_to_copy), Qt.TextFormat.RichText)
        else:
            copy_to_clipboard(self.settings.line_end.join(text_to_copy), Qt.TextFormat.PlainText)

    @Slot()
    def _on_action_copy_current_triggered(self) -> None:
        self.copy_selected_items(self.results_table.selectionModel().currentIndex().column())

    @Slot()
    def _on_action_copy_name_triggered(self) -> None:
        self.copy_selected_items(0)

    @Slot()
    def _on_action_copy_frequency_triggered(self) -> None:
        self.copy_selected_items(1)

    @Slot()
    def _on_action_copy_intensity_triggered(self) -> None:
        self.copy_selected_items(2)

    @Slot()
    def _on_action_copy_lower_state_energy_triggered(self) -> None:
        self.copy_selected_items(3)

    @Slot()
    def _on_action_copy_triggered(self) -> None:
        copy_to_clipboard(self.stringify_selection_html(), Qt.TextFormat.RichText)

    @Slot()
    def _on_action_select_all_triggered(self) -> None:
        self.results_table.selectAll()

    @Slot()
    def _on_action_substance_info_triggered(self) -> None:
        if self.results_table.selectionModel().hasSelection():
            syn: SubstanceInfo = SubstanceInfo(
                self.catalog,
                self.results_model.row(self.results_table.selectionModel().selectedRows()[0].row()).species_tag,
                inchi_key_search_url_template=self.settings.inchi_key_search_url_template,
                parent=self,
            )
            syn.exec()

    def toggle_results_table_column_visibility(self, column: int, is_visible: bool) -> None:
        if is_visible != self.results_table.isColumnHidden(column):
            return
        if is_visible:
            self.results_table.showColumn(column)
        else:
            self.results_table.hideColumn(column)

    @Slot(bool)
    def _on_action_show_substance_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(0, is_checked)

    @Slot(bool)
    def _on_action_show_frequency_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(1, is_checked)

    @Slot(bool)
    def _on_action_show_intensity_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(2, is_checked)

    @Slot(bool)
    def _on_action_show_lower_state_energy_toggled(self, is_checked: bool) -> None:
        self.toggle_results_table_column_visibility(3, is_checked)

    @Slot()
    def _on_action_check_updates_triggered(self) -> None:
        _latest_release: ReleaseInfo = latest_release()
        if not _latest_release:
            QMessageBox.warning(self, self.tr("Release Info"), self.tr("Update check failed."))
        elif _latest_release.version > __version__:
            res: QMessageBox.StandardButton = QMessageBox.question(
                self,
                self.tr("Release Info"),
                self.tr(
                    "Version {release.version} published {release.pub_date} is available. "
                    "Would you like to get the update? "
                    "The app will try to restart."
                ).format(release=_latest_release),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No | QMessageBox.StandardButton.Ignore,
            )
            if res == QMessageBox.StandardButton.Yes:
                update_with_pip()
            elif res == QMessageBox.StandardButton.Ignore:
                self.settings.ignored_version = _latest_release.version
        else:
            QMessageBox.information(self, self.tr("Release Info"), self.tr("You are using the latest version."))

    @Slot()
    def _on_action_about_catalogs_triggered(self) -> None:
        if self.catalog:
            ci: CatalogInfo = CatalogInfo(settings=self.settings, catalog=self.catalog, parent=self)
            ci.catalogUpdated.connect(self._on_action_reload_triggered)
            ci.exec()
        else:
            QMessageBox.information(self, self.tr("Catalog Info"), self.tr("No catalogs loaded"))

    @Slot()
    def _on_action_about_triggered(self) -> None:
        lines: list[str] = [
            self.tr(
                "CatSearch is a means of searching through spectroscopy lines catalogs. It's an offline application."
            ),
            self.tr("It relies on the data stored in JSON files."),
            self.tr(
                "One can use their own catalogs as well as download data from "
                '<a href="https://spec.jpl.nasa.gov/">JPL</a> and '
                '<a href="https://cdms.astro.uni-koeln.de/">CDMS</a> spectroscopy databases '
                "available on the Internet."
            ),
            self.tr("Both plain text JSON and GZip/BZip2/LZMA-compressed JSON are supported."),
            self.tr("See {readme_link} for more info.").format(
                readme_link=a_tag(
                    url="https://github.com/StSav012/pycatsearch/blob/master/README.md", text=self.tr("readme")
                )
            ),
            self.tr("CatSearch is licensed under the {license_link}.").format(
                license_link=a_tag(url="https://www.gnu.org/copyleft/lesser.html", text=self.tr("GNU LGPL version 3"))
            ),
            self.tr("The source code is available on {repo_link}.").format(
                repo_link=a_tag(url="https://github.com/StSav012/pycatsearch", text="GitHub")
            ),
        ]
        about(
            self,
            self.tr("About CatSearch"),
            tag("html", "".join(map(p_tag, lines))),
        )

    @Slot()
    def _on_action_about_qt_triggered(self) -> None:
        QMessageBox.aboutQt(self)

    def load_settings(self) -> None:
        with self.settings.section("search"):
            self.temperature = self.settings.value("temperature", self.spin_temperature.value(), float)
            self.minimal_intensity = self.settings.value("intensity", self.spin_intensity.value(), float)

        with self.settings.section("displayedColumns"):
            for column, (key, action) in enumerate(
                zip(
                    ["substance", "frequency", "intensity", "lowerStateEnergy"],
                    [
                        self.menu_bar.action_show_substance,
                        self.menu_bar.action_show_frequency,
                        self.menu_bar.action_show_intensity,
                        self.menu_bar.action_show_lower_state_energy,
                    ],
                    strict=True,
                )
            ):
                is_visible: bool = self.settings.value(key, True, bool)
                action.setChecked(is_visible)
                self.toggle_results_table_column_visibility(column, is_visible)

        # Fallback: Center the window
        screen: QScreen = QApplication.primaryScreen()
        self.move(
            round(0.5 * (screen.size().width() - self.size().width())),
            round(0.5 * (screen.size().height() - self.size().height())),
        )

        self.settings.restore(self)
        self.settings.restore(self._top_matter)
        self.settings.restore(self._central_widget)
        self.settings.restore(self.results_table.horizontalHeader())
        self.fill_parameters()

        if self.settings.load_last_catalogs:
            self.load_catalog(*self.settings.catalog_file_names)

    def save_settings(self) -> None:
        self.settings.catalog_file_names = self.catalog.sources
        with self.settings.section("search"):
            self.settings.setValue("temperature", self.temperature)
            self.settings.setValue("intensity", self.minimal_intensity)

        with self.settings.section("displayedColumns"):
            self.settings.setValue("substance", self.menu_bar.action_show_substance.isChecked())
            self.settings.setValue("frequency", self.menu_bar.action_show_frequency.isChecked())
            self.settings.setValue("intensity", self.menu_bar.action_show_intensity.isChecked())
            self.settings.setValue("lowerStateEnergy", self.menu_bar.action_show_lower_state_energy.isChecked())

        self.settings.save(self)
        self.settings.save(self._top_matter)
        self.settings.save(self._central_widget)
        self.settings.save(self.results_table.horizontalHeader())

        self.box_substance.save_settings()
        self.box_frequency.save_settings()
        self.settings.sync()

    def preset_table(self) -> None:
        self.results_table.clearSelection()
        self.menu_bar.action_copy.setDisabled(True)
        self.menu_bar.action_substance_info.setDisabled(True)
        self.menu_bar.action_select_all.setDisabled(True)
        self.menu_bar.action_clear.setDisabled(True)
        self.menu_bar.menu_copy_only.setDisabled(True)
        self.results_model.update_units()
        self.update()

    def fill_parameters(self) -> None:
        # frequency
        if not self.catalog.is_empty:
            self.box_frequency.set_frequency_limits(self.catalog.min_frequency, self.catalog.max_frequency)
        self.box_frequency.fill_parameters()

        # intensity
        self.spin_intensity.setSuffix(" " + self.settings.intensity_unit_str)
        self.spin_intensity.setValue(self.settings.from_log10_sq_nm_mhz(self.minimal_intensity))

        # temperature
        temperature_suffix: int = self.settings.temperature_unit
        self.spin_temperature.setSuffix(" " + self.settings.TEMPERATURE_UNITS[temperature_suffix])
        if temperature_suffix == 0:  # K
            self.spin_temperature.setValue(self.temperature)
            self.spin_temperature.setMinimum(0.0)
        elif temperature_suffix == 1:  # °C
            self.spin_temperature.setMinimum(-273.15)
            self.spin_temperature.setValue(self.settings.from_k(self.temperature))
        else:
            raise IndexError("Wrong temperature unit index", temperature_suffix)

    def fill_table(self) -> None:
        self.preset_table()

        if self.box_substance.isChecked() and not self.box_substance.selected_substances:
            self.results_model.clear()
            return

        self.results_table.setSortingEnabled(False)

        entries: CatalogType = self.catalog.filter_by_species_tags(
            species_tags=self.box_substance.selected_substances if self.box_substance.isChecked() else None,
            min_frequency=self.box_frequency.min_frequency,
            max_frequency=self.box_frequency.max_frequency,
            min_intensity=self.minimal_intensity,
            temperature=self.temperature,
        )
        self.results_model.set_entries(entries)

        self.results_table.setSortingEnabled(True)
        self.menu_bar.action_select_all.setEnabled(bool(entries))
        self.menu_bar.action_clear.setEnabled(bool(entries))
        self.menu_bar.menu_copy_only.setEnabled(bool(entries))

    @Slot()
    def _on_search_requested(self) -> None:
        self.status_bar.showMessage(self.tr("Searching…"))
        self.setDisabled(True)
        last_cursor: QCursor = self.cursor()
        self.setCursor(Qt.CursorShape.WaitCursor)
        self.repaint()
        self.fill_table()
        self.setCursor(last_cursor)
        self.setEnabled(True)
        self.status_bar.showMessage(self.tr("Ready."))



from typing import final

from qtpy.QtCore import QCoreApplication
from qtpy.QtWidgets import QWidget

from .file_dialog import OpenFileDialog, SaveFileDialog
from .settings import Settings

__all__ = ["CatalogOpenFileDialog", "CatalogSaveFileDialog"]

_translate = QCoreApplication.translate

supported_name_filters = [
    OpenFileDialog.SupportedNameFilterItem(
        required_packages=["gzip"],
        name=_translate("file type", "JSON with GZip compression"),
        file_extensions=[".json.gz"],
    ),
    OpenFileDialog.SupportedNameFilterItem(
        required_packages=["bz2"],
        name=_translate("file type", "JSON with Bzip2 compression"),
        file_extensions=[".json.bz2"],
    ),
    OpenFileDialog.SupportedNameFilterItem(
        required_packages=["lzma"],
        name=_translate("file type", "JSON with LZMA2 compression"),
        file_extensions=[".json.xz", ".json.lzma"],
    ),
]

supported_mimetype_filters = [
    OpenFileDialog.SupportedMimetypeItem(
        required_packages=[],
        file_extension=".json",
    ),
]


@final
class CatalogOpenFileDialog(OpenFileDialog):
    def __init__(
        self,
        settings: Settings,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(
            settings=settings,
            supported_name_filters=supported_name_filters,
            supported_mimetype_filters=supported_mimetype_filters,
            parent=parent,
        )


@final
class CatalogSaveFileDialog(SaveFileDialog):
    def __init__(
        self,
        settings: Settings,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(
            settings=settings,
            supported_name_filters=supported_name_filters,
            supported_mimetype_filters=supported_mimetype_filters,
            parent=parent,
        )



import mimetypes
import re
from importlib.util import find_spec
from os import PathLike
from pathlib import Path
from typing import Collection, NamedTuple, Sequence, cast

from qtpy.QtCore import QObject, Slot
from qtpy.QtWidgets import QComboBox, QFileDialog, QWidget

from .settings import Settings

__all__ = ["OpenFileDialog", "SaveFileDialog"]


def remove_suffixes(string: str, suffixes: Collection[str]) -> str:
    for suf in sorted(suffixes, key=lambda s: (len(s), s), reverse=True):
        if string.endswith(suf):
            return string[: -len(suf)]
    return string


# noinspection PyPep8Naming
class FileDialog(QFileDialog):
    class SupportedMimetypeItem(NamedTuple):
        required_packages: Collection[str]
        file_extension: str

    class SupportedNameFilterItem(NamedTuple):
        required_packages: Collection[str]
        name: str
        file_extensions: Collection[str]

    def __init__(
        self,
        settings: Settings,
        supported_mimetype_filters: Collection[SupportedMimetypeItem] = (),
        supported_name_filters: Collection[SupportedNameFilterItem] = (),
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent=parent)

        self.settings: Settings = settings
        self.supported_mimetype_filters: Collection[FileDialog.SupportedMimetypeItem] = tuple(
            supported_mimetype_filters
        )
        self.supported_name_filters: Collection[FileDialog.SupportedNameFilterItem] = tuple(supported_name_filters)

        self._selected_filter: str = ""

        @Slot(str)
        def on_filter_selected(selected_filter: str) -> None:
            new_selected_filter_suffixes: tuple[str, ...] = self._extensions_from_filter(selected_filter)
            if not new_selected_filter_suffixes:
                return
            selected_filter_suffixes: tuple[str, ...] = self._extensions_from_filter(self._selected_filter)
            selected_file: Path | None = self.selectedFile()
            if selected_file is None:
                return
            if not selected_file.name.endswith(new_selected_filter_suffixes):
                new_selected_filter_suffix: str = new_selected_filter_suffixes[0]
                new_selected_file = selected_file.with_name(
                    remove_suffixes(selected_file.name, selected_filter_suffixes) + new_selected_filter_suffix
                )
                self.selectFile(new_selected_file)
            self._selected_filter = selected_filter

        def find_widget(name: str) -> QObject | None:
            for child in self.children():
                if child.objectName() == name:
                    return child

        fileTypeCombo: QObject | None = find_widget("fileTypeCombo")
        if fileTypeCombo is not None:
            cast(QComboBox, fileTypeCombo).activated.disconnect()
        self.filterSelected.connect(on_filter_selected)

    def selectFile(self, filename: str | PathLike[str]) -> None:
        return super().selectFile(str(filename))

    def selectedFile(self) -> Path | None:
        try:
            return Path(self.selectedFiles()[0])
        except IndexError:
            return None

    def setNameFilters(self, filters: Sequence[str]) -> None:
        super().setNameFilters(filters)
        if filters:
            self._selected_filter = filters[0]

    def _extensions_from_filter(self, filter_string: str) -> tuple[str, ...]:
        if self.testOption(QFileDialog.Option.HideNameFilterDetails):
            for f in self.nameFilters():
                if f.startswith(filter_string):
                    filter_string = f
                    break
        m: re.Match[str] | None = re.fullmatch(
            r"^.*\(" + r"(?:\*(\.\S+)\s*)" * filter_string.count("*.") + r"\)$",
            filter_string,
        )
        if m is None:
            return ()
        return m.groups()


class OpenFileDialog(FileDialog):
    def __init__(
        self,
        settings: Settings,
        supported_mimetype_filters: Collection[FileDialog.SupportedMimetypeItem] = (),
        supported_name_filters: Collection[FileDialog.SupportedNameFilterItem] = (),
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(settings, supported_mimetype_filters, supported_name_filters, parent)

        self.setObjectName("openFileDialog")

    def _fill_filters(self) -> None:
        mimetypes.init()

        _space_before_extensions: str = " " * (not self.testOption(QFileDialog.Option.HideNameFilterDetails))

        supported_name_filters_names: list[str] = []
        supported_name_filters: list[OpenFileDialog.SupportedNameFilterItem] = []
        for supported_name_filter in self.supported_name_filters:
            if not supported_name_filter.required_packages or any(
                find_spec(package) for package in supported_name_filter.required_packages
            ):
                supported_name_filters_names.append(
                    "".join(
                        (
                            supported_name_filter.name,
                            _space_before_extensions,
                            "(",
                            " ".join("*" + ext for ext in supported_name_filter.file_extensions),
                            ")",
                        )
                    )
                )
                supported_name_filters.append(supported_name_filter)

        supported_mimetypes: list[str] = []
        mimetype: str | None
        for supported_mimetype_filter in self.supported_mimetype_filters:
            if not supported_mimetype_filter.required_packages or any(
                find_spec(package) for package in supported_mimetype_filter.required_packages
            ):
                if mimetype := mimetypes.types_map.get(supported_mimetype_filter.file_extension):
                    supported_mimetypes.append(mimetype)
        # for the “All files (*)” filter
        supported_mimetypes.append("application/octet-stream")

        self.setMimeTypeFilters(supported_mimetypes)
        all_extensions: list[str] = [
            "*" + ext for t in supported_mimetypes[:-1] for ext in mimetypes.guess_all_extensions(t, strict=False)
        ] + ["*" + ext for t in supported_name_filters for ext in t.file_extensions]
        name_filters: list[str] = supported_name_filters_names + self.nameFilters()
        all_extensions_unique: frozenset[str] = frozenset(all_extensions)
        if len(all_extensions_unique) > 1:
            name_filters.insert(
                0,
                "".join(
                    (self.tr("All supported"), _space_before_extensions, "(", " ".join(all_extensions_unique), ")")
                ),
            )
            self.setDefaultSuffix(all_extensions[0])
        self.setNameFilters(name_filters)

    def get_open_filename(self) -> Path | None:
        self._fill_filters()

        opened_filename: Path | None
        if opened_filename := self.settings.opened_file_name:
            self.selectFile(opened_filename)
            self.setDirectory(str(opened_filename.parent))

        self.settings.restore(self)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        self.setFileMode(QFileDialog.FileMode.ExistingFile)
        if self.exec() and (file_path := self.selectedFile()):
            self.settings.save(self)
            self.settings.opened_file_name = file_path
            return file_path
        return None

    def get_open_filenames(self) -> list[str | PathLike[str]]:
        self._fill_filters()

        opened_filename: Path | None
        if opened_filename := self.settings.opened_file_name:
            self.selectFile(opened_filename)
            self.setDirectory(str(opened_filename.parent))

        self.settings.restore(self)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        self.setFileMode(QFileDialog.FileMode.ExistingFiles)
        if self.exec() and (file_paths := self.selectedFiles()):
            self.settings.save(self)
            self.settings.opened_file_name = file_paths[-1]
            return file_paths
        return []


class SaveFileDialog(FileDialog):
    def __init__(
        self,
        settings: Settings,
        supported_mimetype_filters: Collection[FileDialog.SupportedMimetypeItem] = (),
        supported_name_filters: Collection[FileDialog.SupportedNameFilterItem] = (),
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(settings, supported_mimetype_filters, supported_name_filters, parent)

        self.setObjectName("saveFileDialog")

    def get_save_filename(self) -> Path | None:
        mimetypes.init()

        _space_before_extensions: str = " " * (not self.testOption(QFileDialog.Option.HideNameFilterDetails))

        filename: Path | None = self.settings.saved_file_name
        opened_filename: Path | None = self.settings.opened_file_name

        filename_mimetype: str | None = (
            mimetypes.guess_type(filename, strict=False)[0] if filename is not None else None
        )
        filename_suffixes: str = "".join(filename.suffixes) if filename is not None else ""

        selected_mimetype: str | None = None
        selected_filter: str | None = None
        selected_ext: str = filename.suffix if filename is not None else ""

        ext: str | None

        supported_name_filters: list[str] = []
        all_extensions: set[str] = set()
        for supported_name_filter in self.supported_name_filters:
            if not supported_name_filter.required_packages or any(
                find_spec(package) for package in supported_name_filter.required_packages
            ):
                filter_: str = "".join(
                    (
                        supported_name_filter.name,
                        _space_before_extensions,
                        "(",
                        " ".join("*" + ext for ext in supported_name_filter.file_extensions),
                        ")",
                    )
                )
                supported_name_filters.append(filter_)
                all_extensions.update(supported_name_filter.file_extensions)
                if any(filename_suffixes.endswith(ext) for ext in supported_name_filter.file_extensions):
                    selected_filter = filter_
                    if not selected_ext and supported_name_filter.file_extensions:
                        selected_ext = list(supported_name_filter.file_extensions)[0]

        supported_mimetypes: list[str] = []
        mimetype: str | None
        for supported_mimetype_filter in self.supported_mimetype_filters:
            if not supported_mimetype_filter.required_packages or any(
                find_spec(package) for package in supported_mimetype_filter.required_packages
            ):
                if mimetype := mimetypes.types_map.get(supported_mimetype_filter.file_extension):
                    supported_mimetypes.append(mimetype)
                    all_extensions.add(supported_mimetype_filter.file_extension)
                    if filename_mimetype is not None and filename_mimetype == mimetype:
                        selected_mimetype = mimetype
                        if not selected_ext and (ext := mimetypes.guess_extension(selected_mimetype, strict=False)):
                            selected_ext = ext

        if not supported_name_filters and not supported_mimetypes:
            return None

        self.setMimeTypeFilters(supported_mimetypes)
        if supported_name_filters:
            self.setNameFilters(supported_name_filters + self.nameFilters())
        self.setOption(QFileDialog.Option.DontConfirmOverwrite, False)
        if selected_filter is not None:
            self.selectNameFilter(selected_filter)
        elif selected_mimetype is not None:
            self.selectMimeTypeFilter(selected_mimetype)

        self.settings.restore(self)
        self.setAcceptMode(QFileDialog.AcceptMode.AcceptSave)
        self.setFileMode(QFileDialog.FileMode.AnyFile)
        if selected_filter := self.selectedNameFilter():
            selected_filter_ext: tuple[str, ...] = self._extensions_from_filter(selected_filter)
            if selected_filter_ext:
                selected_ext = selected_filter_ext[0]
                self.setDefaultSuffix(selected_ext)

        expected_file: Path
        if expected_file := (filename or opened_filename):
            self.setDirectory(str(expected_file.parent))
            if opened_filename:
                expected_file = expected_file.with_name(opened_filename.name)
            if selected_ext:
                expected_file = expected_file.with_name(
                    remove_suffixes(expected_file.name, all_extensions) + selected_ext
                )
            self.selectFile(expected_file)

        if self.exec() and self.selectedFiles():
            self.settings.save(self)
            if not (filename := self.selectedFile()):
                return None
            self.settings.saved_file_name = filename
            return filename
        return None



from queue import Queue
from typing import cast

from qtpy.QtCore import QTimer, Slot
from qtpy.QtWidgets import QProgressBar, QVBoxLayout, QWidget, QWizard, QWizardPage

from . import SaveCatalogWizard
from ...catalog import Catalog

try:
    from ...async_downloader import Downloader
except (SyntaxError, ImportError, ModuleNotFoundError):
    from ...downloader import Downloader

__all__ = ["ProgressPage"]


class ProgressPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None, existing_catalog: Catalog | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr("Downloading catalog"))
        self.setCommitPage(True)

        layout: QVBoxLayout = QVBoxLayout(self)
        self._downloader: Downloader | None = None
        self._state_queue: Queue[tuple[int, int]] = Queue()

        self._existing_catalog: Catalog | None = existing_catalog

        self._progress_bar: QProgressBar = QProgressBar(self)
        self._progress_bar.setFormat(
            self.tr(
                "%p% (%v out of %m)",
                "%p = the percentage. %v = the current value. %m = the total number of steps.",
            )
        )
        layout.addWidget(self._progress_bar)

        self._timer: QTimer = QTimer(self)
        self._timer.timeout.connect(self._on_timeout)

    def initializePage(self) -> None:
        super(ProgressPage, self).initializePage()
        self.setButtonText(QWizard.WizardButton.CommitButton, self.buttonText(QWizard.WizardButton.NextButton))

        frequency_limits: tuple[float, float] = (
            self.field("min_frequency"),
            self.field("max_frequency"),
        )
        self._downloader = Downloader(
            frequency_limits=frequency_limits,
            state_queue=self._state_queue,
            existing_catalog=self._existing_catalog,
        )
        self._downloader.start()
        self._progress_bar.setMaximum(0)
        self._timer.start(100)

    @Slot()
    def _on_timeout(self) -> None:
        cataloged_species: int
        not_yet_processed_species: int
        while not self._state_queue.empty():
            cataloged_species, not_yet_processed_species = self._state_queue.get(block=False)
            self._progress_bar.setValue(cataloged_species)
            self._progress_bar.setMaximum(cataloged_species + not_yet_processed_species)
        if self.isComplete():
            self.stop()
            self._timer.stop()
            cast(SaveCatalogWizard, self.wizard()).catalog = self._downloader.catalog
            self.completeChanged.emit()

    def isActive(self) -> bool:
        return self._downloader is not None and self._downloader.is_alive()

    def isComplete(self) -> bool:
        return self._downloader is not None and not self._downloader.is_alive()

    def stop(self) -> None:
        self._downloader.stop()
        cataloged_species: int
        not_yet_processed_species: int
        while not self._state_queue.empty():
            cataloged_species, not_yet_processed_species = self._state_queue.get(block=False)
            self._progress_bar.setValue(cataloged_species)
            self._progress_bar.setMaximum(cataloged_species + not_yet_processed_species)
        self._downloader.join()



"""
Download the species list and save it to ./species.json
and to ./src/pycatsearch/species.json.gz if the ./src/pycatsearch/ path exists.

Backup the existing files.

The JSON is prepended with the copyright notice and appended with the timestamp it was received at.

The code will crash unless the operations are permitted.
"""

import gzip
import json
import logging
import random
import time
from datetime import datetime, timezone
from http import HTTPMethod, HTTPStatus
from http.client import HTTPConnection, HTTPResponse, HTTPSConnection
from pathlib import Path
from typing import Any, Mapping
from urllib.parse import ParseResult, urlencode, urlparse

from pycatsearch.utils import SPECIES_TAG

logger: logging.Logger = logging.getLogger(__file__)


def session_for_url(scheme: str, location: str) -> HTTPConnection | HTTPSConnection:
    if scheme == "http":
        return HTTPConnection(location)
    elif scheme == "https":
        return HTTPSConnection(location)
    else:
        raise ValueError(f"Unknown scheme: {scheme}")


def post(url: str, data: dict[str, Any], headers: Mapping[str, str] | None = None) -> tuple[bytes, str]:
    parse_result: ParseResult = urlparse(url)
    session: HTTPConnection | HTTPSConnection = session_for_url(parse_result.scheme, parse_result.netloc)
    response: HTTPResponse
    while True:
        try:
            session.request(
                method=HTTPMethod.POST,
                url=parse_result.path,
                body=urlencode(data),
                headers=(headers or dict()),
            )
            response = session.getresponse()
        except ConnectionResetError:
            time.sleep(random.random())
        else:
            break
    if response.closed:
        logger.error(f"Stream closed before read the response from {url}")
        return b"", ""
    if response.status != HTTPStatus.OK:
        logger.error(f"Status {response.status} ({response.reason}) while posting to {url}")
        return b"", response.getheader("Date", datetime.now(tz=timezone.utc).isoformat())
    try:
        return response.read(), response.getheader("Date", datetime.now(tz=timezone.utc).isoformat())
    except AttributeError:
        logger.warning("`response.fp` became `None` before the socket began closing")
        return b"", ""


def main() -> None:
    species_list_data: bytes
    timestamp: str
    species_list_data, timestamp = post(
        "https://cdms.astro.uni-koeln.de/cdms/portal/json_list/species/",
        {"database": -1},
        headers={"Content-Type": "application/x-www-form-urlencoded"},
    )
    if not species_list_data:
        print("got no data, exiting")
        return

    data: dict[str, int | str | list[dict[str, None | int | str]]] = json.loads(species_list_data)

    data = {
        "credit": {
            "copyright": "I. Physikalisches Institut der Universität zu Köln",
            "url": "https://cdms.astro.uni-koeln.de/",
            "references": [
                "https://doi.org/10.1016/j.jms.2016.03.005",
                "https://doi.org/10.1016/j.molstruc.2005.01.027",
                "https://doi.org/10.1051/0004-6361:20010367",
            ],
        },
        **data,
        "timestamp": timestamp,
    }

    file: Path = Path("species.json")

    prev_data: dict[str, int | str | list[dict[str, None | int | str]]] = {}
    if file.exists():
        prev_data = json.loads(file.read_text(encoding="utf-8"))
        file.rename(file.with_stem("~" + file.stem))
    species_list_text: str = json.dumps(data, indent=2)
    file.write_text(species_list_text, encoding="utf-8")

    print("saved", len(data.get("species", 0)), "species")
    if prev_data:
        current_tags: set[int] = {s[SPECIES_TAG] for s in data.get("species")}
        prev_tags: set[int] = {s[SPECIES_TAG] for s in prev_data.get("species")}
        print("removed tags:", prev_tags - current_tags)
        print("new tags:", current_tags - prev_tags)

    gzip_file: Path = Path("src") / "pycatsearch" / "species.json.gz"
    if gzip_file.exists():
        gzip_file.rename(gzip_file.with_stem("~" + gzip_file.stem))

    if gzip_file.parent.exists():
        # do not save the archive if the directory tree is missing
        with gzip.open(gzip_file, "wt", encoding="utf-8") as f_out:
            f_out.write(species_list_text)
    else:
        print("skipping", gzip_file)


if __name__ == "__main__":
    main()



import asyncio
import logging
import random
from contextlib import suppress
from math import inf
from pathlib import Path
from platform import system
from queue import Empty, Queue
from ssl import SSLCertVerificationError
from threading import Thread
from typing import Any, Final, Mapping, cast
from urllib.error import HTTPError
from urllib.parse import urlencode

import aiohttp
import aiohttp.client_exceptions

try:
    import orjson as json
except ImportError:
    import json

from .catalog import Catalog, CatalogEntryType, CatalogType
from .catalog_entry import CatalogEntry
from .utils import FREQUENCY, LINES, SPECIES_TAG, DEGREES_OF_FREEDOM, VERSION, within, save_catalog_to_file

__all__ = ["Downloader", "get_catalog", "save_catalog", "download"]

logger: logging.Logger = logging.getLogger("async_downloader")


class Downloader(Thread):
    def __init__(
        self,
        frequency_limits: tuple[float, float] = (-inf, inf),
        *,
        existing_catalog: Catalog | None = None,
        state_queue: Queue[tuple[int, int]] | None = None,
    ) -> None:
        super().__init__()
        self._state_queue: Queue[tuple[int, int]] | None = state_queue
        self._frequency_limits: tuple[float, float] = frequency_limits
        self._catalog: CatalogType = dict()
        self._existing_catalog: Catalog | None = existing_catalog

        self._run: bool = False
        self._tasks: list[asyncio.Task] = []

    def __del__(self) -> None:
        self.stop()

    @property
    def catalog(self) -> CatalogType:
        return self._catalog.copy()

    def stop(self) -> None:
        self._run = False

    def join(self, timeout: float | None = None) -> None:
        self.stop()
        self.cancel()
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        loop.stop()
        loop.close()
        super().join(timeout=timeout)

    def cancel(self) -> None:
        for task in self._tasks:
            if not task.done():
                task.cancel()

    def run(self) -> None:
        self._run = True

        async def async_get_catalog() -> CatalogType:
            session: aiohttp.ClientSession
            async with aiohttp.ClientSession(
                timeout=aiohttp.ClientTimeout(),  # disable timeout checks
                trust_env=True,
            ) as session:

                async def get(url: str, headers: Mapping[str, str] | None = None) -> bytes:
                    ssl: bool | None = None
                    response: aiohttp.ClientResponse
                    while self._run:
                        try:
                            async with session.get(url, headers=headers, ssl=ssl) as response:
                                return await response.read()
                        except asyncio.exceptions.CancelledError as ex:
                            if self._run:
                                logger.error(str(ex), exc_info=ex)
                        except aiohttp.client_exceptions.ServerDisconnectedError as ex:
                            logger.warning(f"{url}: {str(ex.message)}")
                        except aiohttp.client_exceptions.ClientConnectorError as ex:
                            logger.warning(f"{str(ex.args[1])} to {url}")
                            if isinstance(ex.args[1], SSLCertVerificationError) and system() == "Windows":
                                logger.critical("Disabling the SSL Certificate validation for the URL!")
                                ssl = False
                        except aiohttp.client_exceptions.ClientOSError as ex:
                            logger.warning(f"{url}: {str(ex)}")
                        except aiohttp.client_exceptions.ClientPayloadError as ex:
                            logger.warning(f"{url}: {str(ex)}")
                        except aiohttp.client_exceptions.ClientError as ex:
                            logger.error(f"{url}: {str(ex)}", exc_info=ex)
                        with suppress(asyncio.exceptions.CancelledError):
                            await asyncio.sleep(random.random())
                    return bytes()

                async def post(url: str, data: dict[str, Any], headers: Mapping[str, str] | None = None) -> bytes:
                    async with session.post(url, data=urlencode(data).encode(), headers=headers) as response:
                        if response.status != 200:
                            logger.error(f"Status {response.status} ({response.reason}) while posting to {url}")
                            return b""
                        return await response.read()

                async def get_species() -> list[dict[str, int | str]]:
                    def purge_null_data(entry: dict[str, None | int | str]) -> dict[str, int | str]:
                        return dict(
                            (key, value)
                            for key, value in entry.items()
                            if value is not None and value not in ("", "None")
                        )

                    def trim_strings(entry: dict[str, None | int | str]) -> dict[str, None | int | str]:
                        key: str
                        for key in entry:
                            if isinstance(entry[key], str):
                                entry[key] = cast(str, entry[key]).strip()
                        return entry

                    def ensure_unique_species_tags(entries: list[dict[str, int | str]]) -> list[dict[str, int | str]]:
                        items_to_delete: set[int] = set()
                        for i in range(len(entries) - 1):
                            for j in range(i + 1, len(entries)):
                                if entries[i][SPECIES_TAG] == entries[j][SPECIES_TAG]:
                                    if entries[i][VERSION] < entries[j][VERSION]:
                                        items_to_delete.add(i)
                                    else:
                                        items_to_delete.add(j)
                        if items_to_delete:
                            return [entries[i] for i in range(len(entries)) if i not in items_to_delete]
                        else:
                            return entries

                    species_list: bytes = await post(
                        "https://cdms.astro.uni-koeln.de/cdms/portal/json_list/species/",
                        {"database": -1},
                        headers={"Content-Type": "application/x-www-form-urlencoded"},
                    )
                    if not species_list:
                        import gzip

                        try:
                            # try using a local copy of the data
                            with gzip.open(Path(__file__).parent / "species.json.gz", "rb") as f_in:
                                species_list = f_in.read()
                        except (OSError, EOFError, Exception):
                            return []

                    data: dict[str, int | str | list[dict[str, None | int | str]]] = json.loads(species_list)
                    return ensure_unique_species_tags(
                        [purge_null_data(trim_strings(s)) for s in data.get("species", [])]
                    )

                async def get_substance_catalog(species_entry: dict[str, int | str]) -> CatalogEntryType:
                    def entry_url(_species_tag: int) -> str:
                        entry_filename: str = f"c{_species_tag:06}.cat"

                        if entry_filename in ("c044009.cat", "c044012.cat"):
                            return ""  # merged with c044004.cat — Brian J. Drouin
                        if _species_tag % 1000 > 500:
                            return "https://cdms.astro.uni-koeln.de/classic/entries/" + entry_filename
                        else:
                            return "https://spec.jpl.nasa.gov/ftp/pub/catalog/" + entry_filename

                    if SPECIES_TAG not in species_entry:
                        # nothing to go on with
                        logger.error(f"{SPECIES_TAG!r} not in the species entry: {species_entry!r}")
                        return dict()

                    if (
                        self._existing_catalog is not None
                        and self._existing_catalog.min_frequency <= min(self._frequency_limits)
                        and self._existing_catalog.max_frequency >= max(self._frequency_limits)
                    ):
                        species_tag: int
                        existing_catalog_entry: CatalogEntryType
                        for species_tag, existing_catalog_entry in self._existing_catalog.catalog.items():
                            if all(
                                existing_catalog_entry.get(key, type(value)()) == value
                                for key, value in species_entry.items()
                            ):
                                logger.debug(f"using existing entry for species tag {species_tag}")
                                _catalog_entry = existing_catalog_entry.copy()
                                _catalog_entry[LINES] = [
                                    _line
                                    for _line in existing_catalog_entry.get(LINES, [])
                                    if within(_line[FREQUENCY], self._frequency_limits)
                                ]
                                return _catalog_entry

                    fn: str = entry_url(cast(int, species_entry[SPECIES_TAG]))
                    if not fn:  # no need to download a file for the species tag
                        logger.debug(f"skipping species tag {species_entry[SPECIES_TAG]}")
                        return dict()
                    try:
                        lines = (await get(fn)).decode().splitlines()
                    except HTTPError as ex:
                        logger.error(fn, exc_info=ex)
                        return dict()
                    catalog_entries = [CatalogEntry(line) for line in lines]
                    if not catalog_entries:
                        if self._run:
                            logger.warning("no entries in the catalog")
                        return dict()
                    return {
                        **species_entry,
                        DEGREES_OF_FREEDOM: catalog_entries[0].degrees_of_freedom,
                        LINES: [
                            _catalog_entry.to_dict()
                            for _catalog_entry in catalog_entries
                            if within(_catalog_entry.frequency, self._frequency_limits)
                        ],
                    }

                species: list[dict[str, int | str]] = await get_species()
                catalog: CatalogType = dict()
                species_count: Final[int] = len(species)
                skipped_count: int = 0
                if self._state_queue is not None:
                    self._state_queue.put((len(catalog), species_count - len(catalog) - skipped_count))
                self._tasks = [asyncio.create_task(get_substance_catalog(_e)) for _e in species]
                catalog_entry: CatalogEntryType
                future_entry: asyncio.Future[CatalogEntryType]
                for future_entry in asyncio.as_completed(self._tasks):
                    catalog_entry = await future_entry
                    if SPECIES_TAG in catalog_entry:
                        catalog[catalog_entry[SPECIES_TAG]] = catalog_entry
                        if self._state_queue is not None:
                            self._state_queue.put((len(catalog), species_count - len(catalog) - skipped_count))
                    else:
                        skipped_count += 1
                        if self._state_queue is not None and self._run:
                            self._state_queue.put((len(catalog), species_count - len(catalog) - skipped_count))

            return catalog

        with suppress(
            RuntimeError,
            asyncio.exceptions.CancelledError,  # it might be “cannot schedule new futures after shutdown”
            asyncio.exceptions.InvalidStateError,  # unfortunate interruption moment
        ):
            self._catalog = asyncio.run(async_get_catalog())


def get_catalog(
    frequency_limits: tuple[float, float] = (-inf, inf),
    *,
    existing_catalog: Catalog | None = None,
) -> CatalogType:
    """
    Download the spectral lines catalog data

    :param tuple frequency_limits: The frequency range of the catalog entries to keep.
    :param Catalog | None existing_catalog: An existing catalog to base the data on.
        If specified, only the entries not presented in it will be downloaded.
    :return: A list of the spectral lines catalog entries.
    """

    state_queue: Queue[tuple[int, int]] = Queue()
    downloader: Downloader = Downloader(
        frequency_limits=frequency_limits,
        state_queue=state_queue,
        existing_catalog=existing_catalog,
    )
    downloader.start()

    cataloged_species: int
    not_yet_processed_species: int
    while downloader.is_alive():
        try:
            cataloged_species, not_yet_processed_species = state_queue.get(block=True, timeout=0.1)
        except Empty:
            continue
        except KeyboardInterrupt:
            downloader.stop()
        else:
            logger.info(f"got {cataloged_species} entries, {not_yet_processed_species} left")

    while downloader.is_alive():
        try:
            cataloged_species, not_yet_processed_species = state_queue.get(block=True, timeout=0.1)
        except Empty:
            continue
        except KeyboardInterrupt:
            downloader.join(0.1)
        else:
            logger.info(f"got {cataloged_species} entries, {not_yet_processed_species} left")

    while not state_queue.empty():
        try:
            cataloged_species, not_yet_processed_species = state_queue.get()
        except KeyboardInterrupt:
            downloader.join(0.1)
        else:
            logger.info(f"got {cataloged_species} entries, {not_yet_processed_species} left")

    downloader.join()

    return downloader.catalog


def save_catalog(
    filename: str,
    frequency_limits: tuple[float, float] = (0, inf),
    *,
    existing_catalog: Catalog | None = None,
) -> bool:
    """
    Download and save the spectral lines catalog data

    :param str filename: The name of the file to save the downloaded catalog to.
        If it ends with an unknown suffix, `'.json.gz'` is appended to it.
    :param tuple frequency_limits: The tuple of the maximal and the minimal frequencies of the lines being stored.
        All the lines outside the specified frequency range are omitted.
    :param Catalog | None existing_catalog: An existing catalog to base the data on.
        If specified, only the entries not presented in it will be downloaded.
    """

    return save_catalog_to_file(
        filename=filename,
        catalog=get_catalog(frequency_limits, existing_catalog=existing_catalog),
        frequency_limits=frequency_limits,
    )


def download() -> None:
    import argparse
    from datetime import datetime
    from pathlib import Path

    from .catalog import Catalog

    ap: argparse.ArgumentParser = argparse.ArgumentParser(
        allow_abbrev=True,
        description="Download JPL and CDMS spectroscopy catalogs for offline search.\n"
        "Find more at https://github.com/StSav012/pycatsearch.",
    )
    ap.add_argument("catalog", type=Path, help="the catalog location to save into (required)")
    ap.add_argument("-fmin", "--min-frequency", type=float, help="the lower frequency [MHz] to take", default=-inf)
    ap.add_argument("-fmax", "--max-frequency", type=float, help="the upper frequency [MHz] to take", default=+inf)
    ap.add_argument("-b", "--base", type=Path, help="an existing catalog to base the data on", default=None)
    args: argparse.Namespace = ap.parse_intermixed_args()

    logging.basicConfig(level=logging.DEBUG)
    logger.info(f"started at {datetime.now()}")
    save_catalog(
        args.catalog,
        (args.min_frequency, args.max_frequency),
        existing_catalog=Catalog(args.base) if args.base is not None else None,
    )
    logger.info(f"finished at {datetime.now()}")



import abc
from pathlib import Path

from qtpy.QtWidgets import QDialog, QMessageBox, QWidget, QWizard

from ..catalog_file_dialog import CatalogSaveFileDialog
from ..save_catalog_waiting_screen import SaveCatalogWaitingScreen
from ..settings import Settings
from ...catalog import CatalogType

__all__ = ["SaveCatalogWizard"]


class SaveCatalogWizard(QWizard):
    def __init__(
        self,
        settings: Settings,
        default_save_location: Path | None = None,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        self.catalog: CatalogType = dict()
        self.default_save_location: Path | None = default_save_location

        self.save_dialog: CatalogSaveFileDialog = CatalogSaveFileDialog(settings=settings, parent=self)

        self.setModal(True)
        if parent is not None:
            self.setWindowIcon(parent.windowIcon())

    @abc.abstractmethod
    def frequency_limits(self) -> tuple[float, float]: ...

    def done(self, exit_code: QDialog.DialogCode) -> None:
        ws: SaveCatalogWaitingScreen
        if exit_code == QDialog.DialogCode.Accepted and self.catalog:
            if self.default_save_location is not None:
                try:
                    ws = SaveCatalogWaitingScreen(
                        self,
                        filename=self.default_save_location,
                        catalog=self.catalog,
                        frequency_limits=self.frequency_limits(),
                    )
                    ws.exec()
                except OSError as ex:
                    QMessageBox.warning(
                        self,
                        self.tr("Unable to save the catalog"),
                        self.tr("Error {exception} occurred while saving {filename}. Try another location.").format(
                            exception=ex,
                            filename=self.default_save_location,
                        ),
                    )
                else:
                    return super(SaveCatalogWizard, self).done(exit_code)

            save_filename: Path | None
            while True:
                if not (save_filename := self.save_dialog.get_save_filename()):
                    return super(SaveCatalogWizard, self).done(QDialog.DialogCode.Rejected)

                try:
                    ws = SaveCatalogWaitingScreen(
                        self,
                        filename=save_filename,
                        catalog=self.catalog,
                        frequency_limits=self.frequency_limits(),
                    )
                    ws.exec()
                except OSError as ex:
                    QMessageBox.warning(
                        self,
                        self.tr("Unable to save the catalog"),
                        self.tr(
                            "Error {exception} occurred while saving {filename}. Try another location.",
                        ).format(exception=ex, filename=save_filename),
                    )
                else:
                    return super(SaveCatalogWizard, self).done(exit_code)

        return super(SaveCatalogWizard, self).done(exit_code)



from pathlib import Path
from typing import ClassVar

from qtpy.QtCore import Signal, Slot
from qtpy.QtWidgets import QFileDialog, QHBoxLayout, QLineEdit, QToolButton, QWidget

__all__ = ["OpenFilePathEntry"]


class OpenFilePathEntry(QWidget):
    changed: ClassVar[Signal] = Signal(Path, name="changed")

    def __init__(
        self,
        initial_file_path: Path | None = None,
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        self._path: Path | None = None

        layout: QHBoxLayout = QHBoxLayout(self)

        self._label: QLineEdit = QLineEdit(self)
        self.path = initial_file_path
        self._label.setReadOnly(True)
        self._label.setMinimumWidth(self._label.height() * 4)
        layout.addWidget(self._label)

        browse_button: QToolButton = QToolButton(self)
        browse_button.setText(self.tr("&Browse…"))
        browse_button.clicked.connect(self._on_browse_button_clicked)
        layout.addWidget(browse_button)

        self._dialog: QFileDialog = QFileDialog(self)
        self._dialog.setAcceptMode(QFileDialog.AcceptMode.AcceptOpen)
        self._dialog.setFileMode(QFileDialog.FileMode.ExistingFile)
        _space_before_extensions: str = " " * (not self._dialog.testOption(QFileDialog.Option.HideNameFilterDetails))
        self._dialog.setNameFilter("".join((self.tr("Translations"), _space_before_extensions, "(*.qm)")))
        self._dialog.setDefaultSuffix(".qm")
        if self._path is not None:
            self._dialog.selectFile(str(self._path))

    @property
    def path(self) -> Path | None:
        return self._path

    @path.setter
    def path(self, path: Path | None) -> None:
        if path is None or not path.is_file():
            self._path = None
            self._label.clear()
            self._label.setToolTip("")
        else:
            self._path = path
            self._label.setText(str(path))
            self._label.setToolTip(str(self._path))

    @Slot()
    def _on_browse_button_clicked(self) -> None:
        if self._dialog.exec() == QFileDialog.DialogCode.Accepted:
            selected_files: list[str] = self._dialog.selectedFiles()
            if selected_files:
                if Path(selected_files[0]) != self._path:
                    self.path = Path(selected_files[0])
                    self.changed.emit(self._path)



from qtpy.QtCore import QLocale
from qtpy.QtWidgets import QLabel, QVBoxLayout, QWidget, QWizard, QWizardPage

__all__ = ["DownloadConfirmationPage"]


class DownloadConfirmationPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        self.setTitle(self.tr("Downloading catalog"))
        self.setCommitPage(True)

        layout: QVBoxLayout = QVBoxLayout(self)
        self._label: QLabel = QLabel(self)
        layout.addWidget(self._label)

    def initializePage(self) -> None:
        super(DownloadConfirmationPage, self).initializePage()
        self.setButtonText(QWizard.WizardButton.CommitButton, self.tr("&Start"))
        self._label.setText(
            self.tr(
                "Click {button_text} to start the download data for {min_frequency} to {max_frequency} MHz."
            ).format(
                button_text=self.buttonText(QWizard.WizardButton.CommitButton).replace("&", ""),
                min_frequency=QLocale().toString(self.field("min_frequency")),
                max_frequency=QLocale().toString(self.field("max_frequency")),
            )
        )



import re
from contextlib import suppress
from typing import Any, Callable

from qtpy.QtCore import QModelIndex, Qt, Signal, Slot
from qtpy.QtGui import QContextMenuEvent, QIcon
from qtpy.QtWidgets import (
    QAbstractItemView,
    QAbstractScrollArea,
    QCheckBox,
    QGroupBox,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QPushButton,
    QStyle,
    QVBoxLayout,
    QWidget,
)

from .html_style_delegate import HTMLDelegate
from .settings import Settings
from .substance_info import SubstanceInfo, SubstanceInfoSelector
from ..catalog import Catalog
from ..utils import (
    CatalogEntryType,
    ISOTOPOLOG,
    NAME,
    STOICHIOMETRIC_FORMULA,
    STRUCTURAL_FORMULA,
    TRIVIAL_NAME,
    best_name,
    remove_html,
)

__all__ = ["SubstanceBox"]


class SubstanceBox(QGroupBox):
    selectedSubstancesChanged: Signal = Signal(name="selectedSubstancesChanged")

    def __init__(self, catalog: Catalog, settings: Settings, parent: QWidget | None = None) -> None:
        from . import qta_icon  # import locally to avoid a circular import

        super().__init__(parent)

        self._catalog: Catalog = catalog
        self._settings: Settings = settings
        self._selected_substances: set[int] = set()

        self._layout_substance: QVBoxLayout = QVBoxLayout(self)
        self._text_substance: QLineEdit = QLineEdit(self)
        self._list_substance: QListWidget = QListWidget(self)
        self._check_keep_selection: QCheckBox = QCheckBox(self)
        self._button_select_none: QPushButton = QPushButton(self)

        self.setCheckable(True)
        self.setTitle(self.tr("Search Only For…"))
        self._text_substance.setClearButtonEnabled(True)
        self._text_substance.setPlaceholderText(self.tr("Filter"))
        self._layout_substance.addWidget(self._text_substance)
        self._list_substance.setEditTriggers(QAbstractItemView.EditTrigger.NoEditTriggers)
        self._list_substance.setDropIndicatorShown(False)
        self._list_substance.setAlternatingRowColors(True)
        self._list_substance.setSelectionMode(QAbstractItemView.SelectionMode.ExtendedSelection)
        self._list_substance.setSelectionBehavior(QAbstractItemView.SelectionBehavior.SelectRows)
        self._list_substance.setSortingEnabled(False)
        self._list_substance.setSizeAdjustPolicy(QAbstractScrollArea.SizeAdjustPolicy.AdjustToContents)
        self._list_substance.setItemDelegateForColumn(0, HTMLDelegate(self._list_substance))
        self._layout_substance.addWidget(self._list_substance)
        self._check_keep_selection.setStatusTip(self.tr("Keep substances list selection through filter changes"))
        self._check_keep_selection.setText(self.tr("Persistent Selection"))
        self._layout_substance.addWidget(self._check_keep_selection)
        self._button_select_none.setStatusTip(self.tr("Clear substances list selection"))
        self._button_select_none.setText(self.tr("Select None"))
        self._layout_substance.addWidget(self._button_select_none)

        self._button_select_none.setIcon(qta_icon("mdi6.checkbox-blank-off-outline"))

        self._text_substance.textChanged.connect(self._on_text_changed)
        self._check_keep_selection.toggled.connect(self._on_check_save_selection_toggled)
        self._button_select_none.clicked.connect(self._on_button_select_none_clicked)
        self._list_substance.doubleClicked.connect(self._on_list_substance_double_clicked)
        self._list_substance.itemChanged.connect(self._on_list_substance_item_changed)

        self.load_settings()

    def contextMenuEvent(self, event: QContextMenuEvent) -> None:
        species_tags: set[int] = self._list_substance.currentItem().data(Qt.ItemDataRole.UserRole)
        if not species_tags:
            return super().contextMenuEvent(event)

        context_menu: QMenu = QMenu(self)
        context_menu.addAction(
            self._icon(
                "dialog-information",
                "mdi6.flask-empty-outline",
                "mdi6.information-variant",
                options=[{}, {"scale_factor": 0.5}],
            ),
            self.tr("Substance &Info") if len(species_tags) == 1 else self.tr("&Select Substance"),
            lambda: self._list_substance.doubleClicked.emit(self._list_substance.currentIndex()),
        )
        context_menu.exec(event.globalPos())
        return super().contextMenuEvent(event)

    def _icon(
        self,
        theme_name: str,
        *qta_name: str,
        standard_pixmap: QStyle.StandardPixmap | None = None,
        **qta_specs: Any,
    ) -> QIcon:
        if theme_name and QIcon.hasThemeIcon(theme_name):
            return QIcon.fromTheme(theme_name)

        if qta_name:
            with suppress(ImportError, Exception):
                import qtawesome as qta

                return qta.icon(*qta_name, **qta_specs)  # might raise an `Exception` if the icon is not in the font

        if standard_pixmap is not None:
            return self.style().standardIcon(standard_pixmap)

        return QIcon()

    def _filter_substances_list(self, filter_text: str) -> dict[str, set[int]]:
        list_items: dict[str, set[int]] = dict()
        allow_html: bool = self._settings.rich_text_in_formulas
        plain_text_name: str
        species_tag: int
        entry: CatalogEntryType
        if filter_text:
            is_filter_regexp: bool = False
            if filter_text.startswith("/"):
                is_filter_regexp = True
                closing_slash_position: int = filter_text.rfind("/")
                pattern: re.Pattern[str]
                try:
                    if closing_slash_position:
                        flag: re.RegexFlag = re.RegexFlag.NOFLAG
                        for f in filter_text[closing_slash_position + 1 :].casefold():
                            flag |= {"a": re.A, "i": re.I, "m": re.M, "s": re.S}.get(f, re.RegexFlag.NOFLAG)
                        pattern = re.compile(filter_text[1:closing_slash_position], flag)
                    else:
                        pattern = re.compile(filter_text[1:])
                except re.error:
                    is_filter_regexp = False
                else:
                    for match_function in (pattern.fullmatch, pattern.match, pattern.search):
                        for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                            for species_tag, entry in self._catalog.catalog.items():
                                with suppress(LookupError):
                                    plain_text_name = remove_html(str(getattr(entry, name_key)))
                                    if match_function(plain_text_name):
                                        if plain_text_name not in list_items:
                                            list_items[plain_text_name] = set()
                                        list_items[plain_text_name].add(species_tag)
                                        html_name = best_name(entry, allow_html=allow_html)
                                        try:
                                            list_items[html_name].add(species_tag)
                                        except LookupError:
                                            list_items[html_name] = {species_tag}
            if not is_filter_regexp:
                filter_text_lowercase: str = filter_text.casefold()
                cmp_function: Callable[[str, str], bool]
                for cmp_function in (str.startswith, str.__contains__):
                    for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                        for species_tag, entry in self._catalog.catalog.items():
                            with suppress(LookupError):
                                plain_text_name = remove_html(str(getattr(entry, name_key)))
                                if cmp_function(plain_text_name, filter_text) or (
                                    name_key in (NAME, TRIVIAL_NAME)
                                    and cmp_function(plain_text_name.casefold(), filter_text_lowercase)
                                ):
                                    if plain_text_name not in list_items:
                                        list_items[plain_text_name] = set()
                                    list_items[plain_text_name].add(species_tag)
                                    html_name = best_name(entry, allow_html=allow_html)
                                    try:
                                        list_items[html_name].add(species_tag)
                                    except LookupError:
                                        list_items[html_name] = {species_tag}
            # species tag suspected
            if filter_text.isdecimal():
                for species_tag in self._catalog.catalog:
                    plain_text_name = str(species_tag)
                    if plain_text_name.startswith(filter_text):
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(species_tag)
            # InChI Key match, see https://en.wikipedia.org/wiki/International_Chemical_Identifier#InChIKey
            if (
                len(filter_text) == 27
                and filter_text[14] == "-"
                and filter_text[25] == "-"
                and filter_text.count("-") == 2
            ):
                for species_tag, entry in self._catalog.catalog.items():
                    plain_text_name = str(entry.inchikey)
                    if plain_text_name == filter_text:
                        if plain_text_name not in list_items:
                            list_items[plain_text_name] = set()
                        list_items[plain_text_name].add(species_tag)
        else:
            for name_key in (ISOTOPOLOG, NAME, STRUCTURAL_FORMULA, STOICHIOMETRIC_FORMULA, TRIVIAL_NAME):
                for species_tag, entry in self._catalog.catalog.items():
                    plain_text_name = remove_html(str(getattr(entry, name_key)))
                    if plain_text_name not in list_items:
                        list_items[plain_text_name] = set()
                    list_items[plain_text_name].add(species_tag)
            list_items = dict(sorted(list_items.items()))
        return list_items

    def _fill_substances_list(self, filter_text: str | None = None) -> None:
        if not filter_text:
            filter_text = self._text_substance.text()

        self._list_substance.clear()

        filtered_items: dict[str, set[int]] = self._filter_substances_list(filter_text)
        text: str
        species_tags: set[int]
        check_state: Qt.CheckState
        for check_state in (Qt.CheckState.Checked, Qt.CheckState.PartiallyChecked, Qt.CheckState.Unchecked):
            for text, species_tags in filtered_items.items():
                new_item_check_state: Qt.CheckState
                if species_tags <= self._selected_substances:
                    new_item_check_state = Qt.CheckState.Checked
                elif species_tags & self._selected_substances:
                    new_item_check_state = Qt.CheckState.PartiallyChecked
                else:
                    new_item_check_state = Qt.CheckState.Unchecked
                if check_state != new_item_check_state:
                    continue
                new_item: QListWidgetItem = QListWidgetItem(text)
                new_item.setData(Qt.ItemDataRole.UserRole, species_tags)
                new_item.setCheckState(new_item_check_state)
                self._list_substance.addItem(new_item)

        if not self._check_keep_selection.isChecked():
            newly_selected_substances: set[int] = set().union(
                *(
                    (self._list_substance.item(row).data(Qt.ItemDataRole.UserRole) & self._selected_substances)
                    for row in range(self._list_substance.count())
                )
            )
            if newly_selected_substances != self._selected_substances:
                self._selected_substances = newly_selected_substances
                self.selectedSubstancesChanged.emit()
        self._text_substance.setFocus()

    @Slot(str)
    def _on_text_changed(self, current_text: str) -> None:
        self._fill_substances_list(current_text)

    @Slot(bool)
    def _on_check_save_selection_toggled(self, new_state: bool) -> None:
        if not new_state:
            newly_selected_substances: set[int] = set().union(
                *(
                    (self._list_substance.item(row).data(Qt.ItemDataRole.UserRole) & self._selected_substances)
                    for row in range(self._list_substance.count())
                )
            )
            if newly_selected_substances != self._selected_substances:
                self._selected_substances = newly_selected_substances
                self.selectedSubstancesChanged.emit()

    @Slot()
    def _on_button_select_none_clicked(self) -> None:
        self._list_substance.blockSignals(True)
        for i in range(self._list_substance.count()):
            self._list_substance.item(i).setCheckState(Qt.CheckState.Unchecked)
        self._list_substance.blockSignals(False)
        self._selected_substances.clear()
        self.selectedSubstancesChanged.emit()

    @Slot(QModelIndex)
    def _on_list_substance_double_clicked(self, index: QModelIndex) -> None:
        @Slot(int, bool)
        def on_tag_selection_changed(species_tag: int, selected: bool) -> None:
            if selected:
                self._selected_substances.add(species_tag)
            else:
                self._selected_substances.discard(species_tag)
            for i in range(self._list_substance.count()):
                _item: QListWidgetItem = self._list_substance.item(i)
                _species_tags: set[int] = _item.data(Qt.ItemDataRole.UserRole)
                new_item_check_state: Qt.CheckState
                if _species_tags <= self._selected_substances:
                    new_item_check_state = Qt.CheckState.Checked
                elif _species_tags & self._selected_substances:
                    new_item_check_state = Qt.CheckState.PartiallyChecked
                else:
                    new_item_check_state = Qt.CheckState.Unchecked
                if _item.checkState() != new_item_check_state:
                    self._list_substance.blockSignals(True)
                    _item.setCheckState(new_item_check_state)
                    self._list_substance.blockSignals(False)
                    self.selectedSubstancesChanged.emit()

        item: QListWidgetItem = self._list_substance.item(index.row())
        species_tags: set[int] = item.data(Qt.ItemDataRole.UserRole).copy()
        if len(species_tags) > 1:
            allow_html: bool = self._settings.rich_text_in_formulas
            sis: SubstanceInfoSelector = SubstanceInfoSelector(
                self.catalog,
                species_tags,
                selected_species_tags=self._selected_substances,
                inchi_key_search_url_template=self._settings.inchi_key_search_url_template,
                allow_html=allow_html,
                parent=self,
            )
            sis.tagSelectionChanged.connect(on_tag_selection_changed)
            sis.exec()
        elif species_tags:  # if not empty
            syn: SubstanceInfo = SubstanceInfo(
                self.catalog,
                species_tags.pop(),
                inchi_key_search_url_template=self._settings.inchi_key_search_url_template,
                parent=self,
            )
            syn.exec()

    @Slot(QListWidgetItem)
    def _on_list_substance_item_changed(self, item: QListWidgetItem) -> None:
        species_tags: set[int] = item.data(Qt.ItemDataRole.UserRole)
        if item.checkState() == Qt.CheckState.Checked:
            if not self._selected_substances.issuperset(species_tags):
                self._selected_substances |= species_tags
                self.selectedSubstancesChanged.emit()
        else:
            if self._selected_substances.intersection(species_tags):
                self._selected_substances -= species_tags
                self.selectedSubstancesChanged.emit()
        self._list_substance.blockSignals(True)
        another_item: QListWidgetItem
        for i in range(self._list_substance.count()):
            another_item = self._list_substance.item(i)
            another_item_species_tags: set[int] = another_item.data(Qt.ItemDataRole.UserRole)
            if another_item_species_tags <= self._selected_substances:
                another_item.setCheckState(Qt.CheckState.Checked)
            elif another_item_species_tags & self._selected_substances:
                another_item.setCheckState(Qt.CheckState.PartiallyChecked)
            else:
                another_item.setCheckState(Qt.CheckState.Unchecked)
        self._list_substance.blockSignals(False)

    def load_settings(self) -> None:
        with self._settings.section("search"), self._settings.section("selection"):
            self._text_substance.setText(self._settings.value("filter", self._text_substance.text(), str))
            self._check_keep_selection.setChecked(self._settings.value("isPersistent", False, bool))
            self.setChecked(self._settings.value("enabled", self.isChecked(), bool))

    def save_settings(self) -> None:
        with self._settings.section("search"), self._settings.section("selection"):
            self._settings.setValue("filter", self._text_substance.text())
            self._settings.setValue("isPersistent", self._check_keep_selection.isChecked())
            self._settings.setValue("enabled", self.isChecked())

    @property
    def catalog(self) -> Catalog:
        return self._catalog

    @catalog.setter
    def catalog(self, new_value: Catalog) -> None:
        self._catalog = new_value
        self._fill_substances_list()

    @property
    def selected_substances(self) -> set[int]:
        if not self.isChecked():
            return set()
        return self._selected_substances
